# Copyright 2014 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
import json

from botocore.model import (
    DenormalizedStructureBuilder,
    ListShape,
    MapShape,
    Shape,
    ShapeResolver,
    StringShape,
    StructureShape,
)

from awscli.arguments import CustomArgument
from awscli.bcdoc.restdoc import ReSTDocument
from awscli.clidocs import (
    CLIDocumentEventHandler,
    GlobalOptionsDocumenter,
    OperationDocumentEventHandler,
    ServiceDocumentEventHandler,
    TopicDocumentEventHandler,
    TopicListerDocumentEventHandler,
)
from awscli.help import (
    HelpCommand,
    ServiceHelpCommand,
    TopicHelpCommand,
    TopicListerCommand,
)
from awscli.testutils import FileCreator, mock, unittest


class TestRecursiveShapes(unittest.TestCase):
    def setUp(self):
        self.arg_table = {}
        self.help_command = mock.Mock()
        self.help_command.event_class = 'custom'
        self.help_command.arg_table = self.arg_table
        self.operation_model = mock.Mock()
        self.operation_model.service_model.operation_names = []
        self.help_command.obj = self.operation_model
        self.operation_handler = OperationDocumentEventHandler(
            self.help_command
        )

    def assert_rendered_docs_contain(self, expected):
        writes = [
            args[0][0] for args in self.help_command.doc.write.call_args_list
        ]
        writes = '\n'.join(writes)
        self.assertIn(expected, writes)

    def assert_proper_indentation(self):
        indent = self.help_command.doc.style.indent.call_count
        dedent = self.help_command.doc.style.dedent.call_count
        message = 'Imbalanced indentation: indent (%s) != dedent (%s)'
        self.assertEqual(indent, dedent, message % (indent, dedent))

    def test_handle_recursive_input(self):
        shape_map = {
            'RecursiveStruct': {
                'type': 'structure',
                'members': {
                    'A': {'shape': 'NonRecursive'},
                    'B': {'shape': 'RecursiveStruct'},
                },
            },
            'NonRecursive': {'type': 'string'},
        }
        shape = StructureShape(
            'RecursiveStruct',
            shape_map['RecursiveStruct'],
            ShapeResolver(shape_map),
        )

        self.arg_table['arg-name'] = mock.Mock(argument_model=shape)
        self.operation_handler.doc_option_example(
            'arg-name', self.help_command, 'process-cli-arg.foo.bar'
        )
        self.assert_rendered_docs_contain('{ ... recursive ... }')

    def test_handle_recursive_output(self):
        shape_map = {
            'RecursiveStruct': {
                'type': 'structure',
                'members': {
                    'A': {'shape': 'NonRecursive'},
                    'B': {'shape': 'RecursiveStruct'},
                },
            },
            'NonRecursive': {'type': 'string'},
        }
        shape = StructureShape(
            'RecursiveStruct',
            shape_map['RecursiveStruct'],
            ShapeResolver(shape_map),
        )

        operation_model = mock.Mock()
        operation_model.output_shape = shape
        self.help_command.obj = operation_model
        self.operation_handler.doc_output(self.help_command, 'event-name')
        self.assert_rendered_docs_contain('( ... recursive ... )')

    def test_handle_empty_nested_struct(self):
        shape_map = {
            'InputStruct': {
                'type': 'structure',
                'members': {
                    'A': {'shape': 'Empty'},
                },
            },
            'Empty': {'type': 'structure', 'members': {}},
        }
        shape = StructureShape(
            'InputStruct', shape_map['InputStruct'], ShapeResolver(shape_map)
        )

        self.arg_table['arg-name'] = mock.Mock(argument_model=shape)
        self.operation_handler.doc_option_example(
            'arg-name', self.help_command, 'process-cli-arg.foo.bar'
        )
        self.assert_proper_indentation()

    def test_handle_no_output_shape(self):
        operation_model = mock.Mock()
        operation_model.output_shape = None
        self.help_command.obj = operation_model
        self.operation_handler.doc_output(self.help_command, 'event-name')
        self.assert_rendered_docs_contain('None')

    def test_handle_memberless_output_shape(self):
        shape_map = {'NoMembers': {'type': 'structure', 'members': {}}}
        shape = StructureShape(
            'NoMembers', shape_map['NoMembers'], ShapeResolver(shape_map)
        )

        operation_model = mock.Mock()
        operation_model.output_shape = shape
        self.help_command.obj = operation_model
        self.operation_handler.doc_output(self.help_command, 'event-name')
        self.assert_rendered_docs_contain('None')


class TestCLIDocumentEventHandler(unittest.TestCase):
    def setUp(self):
        self.session = mock.Mock()
        self.obj = None
        self.command_table = {}
        self.arg_table = {}
        self.name = 'my-command'
        self.event_class = 'aws'

    def create_help_command(self):
        help_command = mock.Mock()
        help_command.doc = ReSTDocument()
        help_command.event_class = 'custom'
        help_command.arg_table = {}
        operation_model = mock.Mock()
        operation_model.documentation = 'description'
        operation_model.service_model.operation_names = []
        help_command.obj = operation_model
        return help_command

    def create_tagged_union_shape(self):
        shape_model = {'type': 'structure', 'union': True, 'members': {}}
        tagged_union = StructureShape('tagged_union', shape_model)
        return tagged_union

    def get_help_docs_for_argument(self, shape):
        arg_table = {'arg-name': mock.Mock(argument_model=shape)}
        help_command = mock.Mock()
        help_command.doc = ReSTDocument()
        help_command.event_class = 'custom'
        help_command.arg_table = arg_table
        operation_model = mock.Mock()
        operation_model.service_model.operation_names = []
        help_command.obj = operation_model
        operation_handler = OperationDocumentEventHandler(help_command)
        operation_handler.doc_option('arg-name', help_command)
        return help_command.doc.getvalue().decode('utf-8')

    def test_breadcrumbs_man(self):
        # Create an arbitrary help command class. This was chosen
        # because it is fairly easy to instantiate.
        help_cmd = ServiceHelpCommand(
            self.session,
            self.obj,
            self.command_table,
            self.arg_table,
            self.name,
            self.event_class,
        )

        doc_handler = CLIDocumentEventHandler(help_cmd)
        doc_handler.doc_breadcrumbs(help_cmd)
        # These should not show up in the man page
        self.assertEqual(help_cmd.doc.getvalue().decode('utf-8'), '')

    def test_breadcrumbs_html(self):
        help_cmd = ServiceHelpCommand(
            self.session,
            self.obj,
            self.command_table,
            self.arg_table,
            self.name,
            self.event_class,
        )
        help_cmd.doc.target = 'html'
        doc_handler = CLIDocumentEventHandler(help_cmd)
        doc_handler.doc_breadcrumbs(help_cmd)
        self.assertEqual(
            help_cmd.doc.getvalue().decode('utf-8'), '[ :ref:`aws <cli:aws>` ]\n\n'
        )

    def test_breadcrumbs_service_command_html(self):
        help_cmd = ServiceHelpCommand(
            self.session,
            self.obj,
            self.command_table,
            self.arg_table,
            self.name,
            'ec2',
        )
        help_cmd.doc.target = 'html'
        doc_handler = CLIDocumentEventHandler(help_cmd)
        doc_handler.doc_breadcrumbs(help_cmd)
        self.assertEqual(
            help_cmd.doc.getvalue().decode('utf-8'), '[ :ref:`aws <cli:aws>` ]\n\n'
        )

    def test_breadcrumbs_operation_command_html(self):
        help_cmd = ServiceHelpCommand(
            self.session,
            self.obj,
            self.command_table,
            self.arg_table,
            self.name,
            'ec2.run-instances',
        )
        help_cmd.doc.target = 'html'
        doc_handler = CLIDocumentEventHandler(help_cmd)
        doc_handler.doc_breadcrumbs(help_cmd)
        self.assertEqual(
            help_cmd.doc.getvalue().decode('utf-8'),
            '[ :ref:`aws <cli:aws>` . :ref:`ec2 <cli:aws ec2>` ]\n\n'
        )

    def test_breadcrumbs_wait_command_html(self):
        help_cmd = ServiceHelpCommand(
            self.session,
            self.obj,
            self.command_table,
            self.arg_table,
            self.name,
            's3api.wait.object-exists',
        )
        help_cmd.doc.target = 'html'
        doc_handler = CLIDocumentEventHandler(help_cmd)
        doc_handler.doc_breadcrumbs(help_cmd)
        self.assertEqual(
            help_cmd.doc.getvalue().decode('utf-8'),
            (
                '[ :ref:`aws <cli:aws>` . :ref:`s3api <cli:aws s3api>`'
                ' . :ref:`wait <cli:aws s3api wait>` ]\n\n'
            )
        )

    def test_documents_json_header_shape(self):
        shape = {
            'type': 'string',
            'jsonvalue': True,
            'location': 'header',
            'locationName': 'X-Amz-Header-Name',
        }
        shape = StringShape('JSONValueArg', shape)
        rendered = self.get_help_docs_for_argument(shape)
        self.assertIn('(JSON)', rendered)

    def test_documents_enum_values(self):
        shape = {'type': 'string', 'enum': ['FOO', 'BAZ']}
        shape = StringShape('EnumArg', shape)
        rendered = self.get_help_docs_for_argument(shape)
        self.assertIn('Possible values', rendered)
        self.assertIn('FOO', rendered)
        self.assertIn('BAZ', rendered)

    def test_documents_recursive_input(self):
        shape_map = {
            'RecursiveStruct': {
                'type': 'structure',
                'members': {
                    'A': {'shape': 'NonRecursive'},
                    'B': {'shape': 'RecursiveStruct'},
                },
            },
            'NonRecursive': {'type': 'string'},
        }
        shape = StructureShape(
            'RecursiveStruct',
            shape_map['RecursiveStruct'],
            ShapeResolver(shape_map),
        )
        rendered = self.get_help_docs_for_argument(shape)
        self.assertIn('( ... recursive ... )', rendered)

    def test_documents_nested_structure(self):
        shape_map = {
            'UpperStructure': {
                'type': 'structure',
                'members': {
                    'A': {'shape': 'NestedStruct'},
                    'B': {'shape': 'NestedStruct'},
                },
            },
            'NestedStruct': {
                'type': 'structure',
                'members': {
                    'Nested_A': {'shape': 'Line'},
                    'Nested_B': {'shape': 'Line'},
                },
            },
            'Line': {'type': 'string'},
        }
        shape = StructureShape(
            'UpperStructure',
            shape_map['UpperStructure'],
            ShapeResolver(shape_map),
        )
        rendered = self.get_help_docs_for_argument(shape)
        self.assertEqual(rendered.count('A -> (structure)'), 1)
        self.assertEqual(rendered.count('B -> (structure)'), 1)
        self.assertEqual(rendered.count('Nested_A -> (string)'), 2)
        self.assertEqual(rendered.count('Nested_B -> (string)'), 2)

    def test_documents_nested_list(self):
        shape_map = {
            'UpperList': {
                'type': 'list',
                'member': {'shape': 'NestedStruct'},
            },
            'NestedStruct': {
                'type': 'structure',
                'members': {
                    'Nested_A': {'shape': 'Line'},
                    'Nested_B': {'shape': 'Line'},
                },
            },
            'Line': {'type': 'string'},
        }
        shape = ListShape(
            'UpperList', shape_map['UpperList'], ShapeResolver(shape_map)
        )
        rendered = self.get_help_docs_for_argument(shape)
        self.assertEqual(rendered.count('(structure)'), 1)
        self.assertEqual(rendered.count('Nested_A -> (string)'), 1)
        self.assertEqual(rendered.count('Nested_B -> (string)'), 1)

    def test_documents_nested_map(self):
        shape_map = {
            'UpperMap': {
                'type': 'map',
                'key': {'shape': 'NestedStruct'},
                'value': {'shape': 'NestedStruct'},
            },
            'NestedStruct': {
                'type': 'structure',
                'members': {
                    'Nested_A': {'shape': 'Line'},
                    'Nested_B': {'shape': 'Line'},
                },
            },
            'Line': {'type': 'string'},
        }
        shape = MapShape(
            'UpperMap', shape_map['UpperMap'], ShapeResolver(shape_map)
        )
        rendered = self.get_help_docs_for_argument(shape)
        self.assertEqual(rendered.count('key -> (structure)'), 1)
        self.assertEqual(rendered.count('value -> (structure)'), 1)
        self.assertEqual(rendered.count('Nested_A -> (string)'), 2)
        self.assertEqual(rendered.count('Nested_B -> (string)'), 2)

    def test_description_only_for_crosslink_manpage(self):
        help_command = self.create_help_command()
        operation_handler = OperationDocumentEventHandler(help_command)
        operation_handler.doc_description(help_command=help_command)
        rendered = help_command.doc.getvalue().decode('utf-8')
        # The links are generated in the "man" mode.
        self.assertIn('See also: AWS API Documentation', rendered)

    def test_includes_webapi_crosslink_in_html(self):
        help_command = self.create_help_command()
        # Configure this for 'html' generation:
        help_command.obj.service_model.metadata = {'uid': 'service-1-2-3'}
        help_command.obj.name = 'myoperation'
        help_command.doc.target = 'html'

        operation_handler = OperationDocumentEventHandler(help_command)
        operation_handler.doc_description(help_command=help_command)
        rendered = help_command.doc.getvalue().decode('utf-8')
        # Should expect an externa link because we're generating html.
        self.assertIn(
            'See also: `AWS API Documentation '
            '<https://docs.aws.amazon.com/goto/'
            'WebAPI/service-1-2-3/myoperation>`_',
            rendered,
        )

    def test_includes_streaming_blob_options(self):
        help_command = self.create_help_command()
        blob_shape = Shape('blob_shape', {'type': 'blob'})
        blob_shape.serialization = {'streaming': True}
        blob_arg = CustomArgument('blob_arg', argument_model=blob_shape)
        help_command.arg_table = {'blob_arg': blob_arg}
        operation_handler = OperationDocumentEventHandler(help_command)
        operation_handler.doc_option(
            arg_name='blob_arg', help_command=help_command
        )
        rendered = help_command.doc.getvalue().decode('utf-8')
        self.assertIn('streaming blob', rendered)

    def test_streaming_blob_comes_after_docstring(self):
        help_command = self.create_help_command()
        blob_shape = Shape('blob_shape', {'type': 'blob'})
        blob_shape.serialization = {'streaming': True}
        blob_arg = CustomArgument(
            name='blob_arg', argument_model=blob_shape, help_text='FooBar'
        )
        help_command.arg_table = {'blob_arg': blob_arg}
        operation_handler = OperationDocumentEventHandler(help_command)
        operation_handler.doc_option(
            arg_name='blob_arg', help_command=help_command
        )
        rendered = help_command.doc.getvalue().decode('utf-8')
        self.assertRegex(rendered, r'FooBar[\s\S]*streaming blob')

    def test_includes_tagged_union_options(self):
        help_command = self.create_help_command()
        tagged_union = self.create_tagged_union_shape()
        arg = CustomArgument(name='tagged_union', argument_model=tagged_union)
        help_command.arg_table = {'tagged_union': arg}
        operation_handler = OperationDocumentEventHandler(help_command)
        operation_handler.doc_option(
            arg_name='tagged_union', help_command=help_command
        )
        rendered = help_command.doc.getvalue().decode('utf-8')
        self.assertIn('(tagged union structure)', rendered)

    def test_tagged_union_comes_after_docstring_options(self):
        help_command = self.create_help_command()
        tagged_union = self.create_tagged_union_shape()
        arg = CustomArgument(
            name='tagged_union',
            argument_model=tagged_union,
            help_text='FooBar',
        )
        help_command.arg_table = {'tagged_union': arg}
        operation_handler = OperationDocumentEventHandler(help_command)
        operation_handler.doc_option(
            arg_name='tagged_union', help_command=help_command
        )
        rendered = help_command.doc.getvalue().decode('utf-8')
        self.assertRegex(rendered, r'FooBar[\s\S]*Tagged Union')

    def test_tagged_union_comes_after_docstring_output(self):
        help_command = self.create_help_command()
        tagged_union = self.create_tagged_union_shape()
        tagged_union.documentation = "FooBar"
        shape = (
            DenormalizedStructureBuilder()
            .with_members(
                {
                    'foo': {
                        'type': 'structure',
                        'union': True,
                        'documentation': 'FooBar',
                        'members': {},
                    }
                }
            )
            .build_model()
        )
        help_command.obj.output_shape = shape
        operation_handler = OperationDocumentEventHandler(help_command)
        operation_handler.doc_output(
            help_command=help_command, event_name='foobar'
        )
        rendered = help_command.doc.getvalue().decode('utf-8')
        self.assertRegex(rendered, r'FooBar[\s\S]*Tagged Union')

    def test_documents_required_parameters(self):
        """Tests that required parameters are correctly documented."""
        shape_map = {
            'ParentStructure': {
                'type': 'structure',
                'required': ['RequiredParameter'],
                'members': {
                    'RequiredParameter': {'shape': 'StringMember'},
                    'OptionalParameter': {'shape': 'StringMember'},
                },
            },
            'StringMember': {'type': 'string'},
        }

        resolver = ShapeResolver(shape_map)
        parent_shape = StructureShape(
            'ParentStructure', shape_map['ParentStructure'], resolver
        )

        rendered = self.get_help_docs_for_argument(parent_shape)

        required_index = rendered.find('RequiredParameter -> (string)')
        optional_index = rendered.find('OptionalParameter -> (string)')

        self.assertIn('[required]', rendered[required_index:optional_index])
        optional_text = rendered[optional_index:]
        self.assertNotIn('[required]', optional_text)

    def test_documents_constraints(self):
        shape = {'type': 'string', 'min': 0, 'max': 10, 'pattern': '.*'}
        shape = StringShape('ConstrainedArg', shape)
        arg = CustomArgument('ConstrainedArg', argument_model=shape)
        help_command = self.create_help_command()
        help_command.arg_table = {'ConstrainedArg': arg}
        operation_handler = OperationDocumentEventHandler(help_command)
        operation_handler.doc_option(
            arg_name='ConstrainedArg', help_command=help_command
        )
        rendered = help_command.doc.getvalue().decode('utf-8')
        self.assertIn('Constraints', rendered)
        self.assertIn('min: ``0``', rendered)
        self.assertIn('max: ``10``', rendered)
        self.assertIn('pattern: ``.*``', rendered)

    def test_meta_description_operation_command_html(self):
        help_cmd = ServiceHelpCommand(
            self.session, self.obj, self.command_table, self.arg_table,
            self.name, 'ec2.run-instances'
        )
        help_cmd.doc.target = 'html'
        doc_handler = OperationDocumentEventHandler(help_cmd)
        doc_handler.doc_meta_description(help_cmd)

        meta_description = help_cmd.doc.getvalue().decode('utf-8')
        self.assertIn(".. meta::\n   :description: ", meta_description)
        self.assertIn('to run the ec2 run-instances command', meta_description)

    def test_meta_description_service_html(self):
        help_cmd = ServiceHelpCommand(
            self.session, self.obj, self.command_table, self.arg_table,
            self.name, 'ec2'
        )
        help_cmd.doc.target = 'html'
        doc_handler = ServiceDocumentEventHandler(help_cmd)
        doc_handler.doc_meta_description(help_cmd)

        meta_description = help_cmd.doc.getvalue().decode('utf-8')
        self.assertIn(".. meta::\n   :description: Learn about the AWS CLI ", meta_description)
        self.assertIn(' ec2 commands', meta_description)


class TestTopicDocumentEventHandlerBase(unittest.TestCase):
    def setUp(self):
        self.session = mock.Mock()
        self.file_creator = FileCreator()

        self.tags_dict = {}

        # Make a temporary json index to base information on
        self.json_index = self.file_creator.create_file('index.json', '')
        with open(self.json_index, 'w') as f:
            json.dump(self.tags_dict, f, indent=4, sort_keys=True)

        self.index_patch = mock.patch(
            'awscli.topictags.TopicTagDB.index_file', self.json_index
        )
        self.dir_patch = mock.patch(
            'awscli.topictags.TopicTagDB.topic_dir', self.file_creator.rootdir
        )
        self.index_patch.start()
        self.dir_patch.start()

    def tearDown(self):
        self.dir_patch.stop()
        self.index_patch.stop()
        self.file_creator.remove_all()


class TestTopicListerDocumentEventHandler(TestTopicDocumentEventHandlerBase):
    def setUp(self):
        super().setUp()
        self.descriptions = [
            'This describes the first topic',
            'This describes the second topic',
            'This describes the third topic',
        ]
        self.tags_dict = {
            'topic-name-1': {
                'title': ['The first topic title'],
                'description': [self.descriptions[0]],
                'category': ['General'],
            },
            'topic-name-2': {
                'title': ['The second topic title'],
                'description': [self.descriptions[1]],
                'category': ['S3'],
            },
            'topic-name-3': {
                'title': ['The third topic title'],
                'description': [self.descriptions[2]],
                'category': ['General'],
            },
        }

        with open(self.json_index, 'w') as f:
            json.dump(self.tags_dict, f, indent=4, sort_keys=True)

        self.cmd = TopicListerCommand(self.session)
        self.doc_handler = TopicListerDocumentEventHandler(self.cmd)

    def test_breadcrumbs(self):
        self.doc_handler.doc_breadcrumbs(self.cmd)
        self.assertEqual(self.cmd.doc.getvalue().decode('utf-8'), '')
        self.cmd.doc.target = 'html'
        self.doc_handler.doc_breadcrumbs(self.cmd)
        self.assertEqual(
            '[ :ref:`aws <cli:aws>` ]', self.cmd.doc.getvalue().decode('utf-8')
        )

    def test_title(self):
        self.doc_handler.doc_title(self.cmd)
        title_contents = self.cmd.doc.getvalue().decode('utf-8')
        self.assertIn(f'.. _cli:aws help {self.cmd.name}:', title_contents)
        self.assertIn('AWS CLI Topic Guide', title_contents)

    def test_description(self):
        self.doc_handler.doc_description(self.cmd)
        self.assertIn(
            'This is the AWS CLI Topic Guide',
            self.cmd.doc.getvalue().decode('utf-8'),
        )

    def test_subitems_start(self):
        ref_output = [
            '-------\nGeneral\n-------',
            (
                f'* topic-name-1: {self.descriptions[0]}\n'
                f'* topic-name-3: {self.descriptions[2]}\n'
            ),
            '--\nS3\n--',
            f'* topic-name-2: {self.descriptions[1]}\n',
        ]

        self.doc_handler.doc_subitems_start(self.cmd)
        contents = self.cmd.doc.getvalue().decode('utf-8')

        for line in ref_output:
            self.assertIn(line, contents)
        # Make sure the toctree is not in the man page
        self.assertNotIn('.. toctree::', contents)

    def test_subitems_start_html(self):
        self.cmd.doc.target = 'html'
        ref_output = [
            '-------\nGeneral\n-------',
            (
                f'* :ref:`topic-name-1 <cli:aws help topic-name-1>`: {self.descriptions[0]}\n'
                f'* :ref:`topic-name-3 <cli:aws help topic-name-3>`: {self.descriptions[2]}\n'
            ),
            '--\nS3\n--',
            (
                f'* :ref:`topic-name-2 <cli:aws help topic-name-2>`: {self.descriptions[1]}\n'
            ),
        ]

        self.doc_handler.doc_subitems_start(self.cmd)
        contents = self.cmd.doc.getvalue().decode('utf-8')

        for line in ref_output:
            self.assertIn(line, contents)
        # Make sure the hidden toctree is in the html
        self.assertIn('.. toctree::', contents)
        self.assertIn(':hidden:', contents)


class TestTopicDocumentEventHandler(TestTopicDocumentEventHandlerBase):
    def setUp(self):
        super().setUp()
        self.name = 'topic-name-1'
        self.title = 'The first topic title'
        self.description = 'This is about the first topic'
        self.category = 'General'
        self.related_command = 'foo'
        self.related_topic = 'topic-name-2'
        self.topic_body = 'Hello World!'

        self.tags_dict = {
            self.name: {
                'title': [self.title],
                'description': [self.description],
                'category': [self.category],
                'related topic': [self.related_topic],
                'related command': [self.related_command],
            }
        }
        with open(self.json_index, 'w') as f:
            json.dump(self.tags_dict, f, indent=4, sort_keys=True)

        self.cmd = TopicHelpCommand(self.session, self.name)
        self.doc_handler = TopicDocumentEventHandler(self.cmd)

    def test_breadcrumbs(self):
        self.doc_handler.doc_breadcrumbs(self.cmd)
        self.assertEqual(self.cmd.doc.getvalue().decode('utf-8'), '')
        self.cmd.doc.target = 'html'
        self.doc_handler.doc_breadcrumbs(self.cmd)
        self.assertEqual(
            '[ :ref:`aws <cli:aws>` . :ref:`topics <cli:aws help topics>` ]',
            self.cmd.doc.getvalue().decode('utf-8'),
        )

    def test_title(self):
        self.doc_handler.doc_title(self.cmd)
        title_contents = self.cmd.doc.getvalue().decode('utf-8')
        self.assertIn(f'.. _cli:aws help {self.name}:', title_contents)
        self.assertIn(self.title, title_contents)

    def test_description(self):
        lines = [
            ':title: ' + self.title,
            ':description: ' + self.description,
            ':category:' + self.category,
            ':related command: ' + self.related_command,
            ':related topic: ' + self.related_topic,
            self.topic_body,
        ]
        body = '\n'.join(lines)
        self.file_creator.create_file(self.name + '.rst', body)
        self.doc_handler.doc_description(self.cmd)
        contents = self.cmd.doc.getvalue().decode('utf-8')
        self.assertIn(self.topic_body, contents)
        self.assertNotIn(':title ' + self.title, contents)

    def test_description_no_tags(self):
        lines = [self.topic_body]
        body = '\n'.join(lines)
        self.file_creator.create_file(self.name + '.rst', body)
        self.doc_handler.doc_description(self.cmd)
        contents = self.cmd.doc.getvalue().decode('utf-8')
        self.assertIn(self.topic_body, contents)

    def test_description_tags_in_body(self):
        lines = [
            ':title: ' + self.title,
            ':description: ' + self.description,
            ':related command: ' + self.related_command,
        ]
        body_lines = [
            ':related_topic: ' + self.related_topic,
            self.topic_body,
            ':foo: bar',
        ]
        body = '\n'.join(lines + body_lines)
        ref_body = '\n'.join(body_lines)
        self.file_creator.create_file(self.name + '.rst', body)
        self.doc_handler.doc_description(self.cmd)
        contents = self.cmd.doc.getvalue().decode('utf-8')
        self.assertIn(ref_body, contents)

    def test_excludes_global_options(self):
        self.doc_handler.doc_global_option(self.cmd)
        global_options = self.cmd.doc.getvalue().decode('utf-8')
        self.assertNotIn('Global Options', global_options)


class TestGlobalOptionsDocumenter(unittest.TestCase):
    def create_help_command(self):
        types = ['blob', 'integer', 'boolean', 'string']
        arg_table = {}
        for t in types:
            name = f'{t}_type'
            help_text = f'This arg type is {t}'
            choices = ['A', 'B', 'C'] if t == 'string' else []
            arg_table[name] = CustomArgument(
                name=name,
                cli_type_name=t,
                help_text=help_text,
                choices=choices,
            )
        help_command = mock.Mock(spec=HelpCommand)
        help_command.arg_table = arg_table
        help_command.doc = ReSTDocument()
        return help_command

    def create_documenter(self):
        return GlobalOptionsDocumenter(self.create_help_command())

    def test_doc_global_options(self):
        documenter = self.create_documenter()
        options = documenter.doc_global_options()
        self.assertIn('``--string_type`` (string)', options)
        self.assertIn('``--integer_type`` (integer)', options)
        self.assertIn('``--boolean_type`` (boolean)', options)
        self.assertIn('``--blob_type`` (blob)', options)
        self.assertIn('*   A', options)
        self.assertIn('*   B', options)
        self.assertIn('*   C', options)

    def test_doc_global_synopsis(self):
        documenter = self.create_documenter()
        synopsis = documenter.doc_global_synopsis()
        self.assertIn('[--string_type <value>]', synopsis)
        self.assertIn('[--integer_type <value>]', synopsis)
        self.assertIn('[--boolean_type]', synopsis)
        self.assertIn('[--blob_type <value>]', synopsis)



#!/usr/bin/env python
# Copyright 2014 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
from dateutil import parser, tz

from tests.functional.s3 import BaseS3TransferCommandTest


class TestLSCommand(BaseS3TransferCommandTest):
    def test_operations_used_in_recursive_list(self):
        time_utc = "2014-01-09T20:45:49.000Z"
        self.parsed_responses = [
            {
                "CommonPrefixes": [],
                "Contents": [
                    {
                        "Key": "foo/bar.txt",
                        "Size": 100,
                        "LastModified": time_utc,
                    }
                ],
            }
        ]
        stdout, _, _ = self.run_cmd(
            's3 ls s3://bucket/ --recursive', expected_rc=0
        )
        call_args = self.operations_called[0][1]
        # We should not be calling the args with any delimiter because we
        # want a recursive listing.
        self.assertEqual(call_args['Prefix'], '')
        self.assertEqual(call_args['Bucket'], 'bucket')
        self.assertNotIn('delimiter', call_args)
        # Time is stored in UTC timezone, but the actual time displayed
        # is specific to your tzinfo, so shift the timezone to your local's.
        time_local = parser.parse(time_utc).astimezone(tz.tzlocal())
        self.assertEqual(
            stdout,
            '%s        100 foo/bar.txt\n'
            % time_local.strftime('%Y-%m-%d %H:%M:%S'),
        )

    def test_errors_out_with_extra_arguments(self):
        stderr = self.run_cmd('s3 ls --extra-argument-foo', expected_rc=252)[1]
        self.assertIn('Unknown options', stderr)
        self.assertIn('--extra-argument-foo', stderr)

    def test_list_buckets_use_page_size(self):
        stdout, _, _ = self.run_cmd('s3 ls --page-size 8', expected_rc=0)
        call_args = self.operations_called[0][1]
        # The page size gets translated to ``MaxBuckets`` in the s3 model
        self.assertEqual(call_args['MaxBuckets'], 8)

    def test_operations_use_page_size(self):
        time_utc = "2014-01-09T20:45:49.000Z"
        self.parsed_responses = [
            {
                "CommonPrefixes": [],
                "Contents": [
                    {
                        "Key": "foo/bar.txt",
                        "Size": 100,
                        "LastModified": time_utc,
                    }
                ],
            }
        ]
        stdout, _, _ = self.run_cmd(
            's3 ls s3://bucket/ --page-size 8', expected_rc=0
        )
        call_args = self.operations_called[0][1]
        # We should not be calling the args with any delimiter because we
        # want a recursive listing.
        self.assertEqual(call_args['Prefix'], '')
        self.assertEqual(call_args['Bucket'], 'bucket')
        # The page size gets translated to ``MaxKeys`` in the s3 model
        self.assertEqual(call_args['MaxKeys'], 8)

    def test_operations_use_page_size_recursive(self):
        time_utc = "2014-01-09T20:45:49.000Z"
        self.parsed_responses = [
            {
                "CommonPrefixes": [],
                "Contents": [
                    {
                        "Key": "foo/bar.txt",
                        "Size": 100,
                        "LastModified": time_utc,
                    }
                ],
            }
        ]
        stdout, _, _ = self.run_cmd(
            's3 ls s3://bucket/ --page-size 8 --recursive', expected_rc=0
        )
        call_args = self.operations_called[0][1]
        # We should not be calling the args with any delimiter because we
        # want a recursive listing.
        self.assertEqual(call_args['Prefix'], '')
        self.assertEqual(call_args['Bucket'], 'bucket')
        # The page size gets translated to ``MaxKeys`` in the s3 model
        self.assertEqual(call_args['MaxKeys'], 8)
        self.assertNotIn('Delimiter', call_args)

    def test_success_rc_has_prefixes_and_objects(self):
        time_utc = "2014-01-09T20:45:49.000Z"
        self.parsed_responses = [
            {
                "CommonPrefixes": [{"Prefix": "foo/"}],
                "Contents": [
                    {
                        "Key": "foo/bar.txt",
                        "Size": 100,
                        "LastModified": time_utc,
                    }
                ],
            }
        ]
        self.run_cmd('s3 ls s3://bucket/foo', expected_rc=0)

    def test_success_rc_has_only_prefixes(self):
        self.parsed_responses = [{"CommonPrefixes": [{"Prefix": "foo/"}]}]
        self.run_cmd('s3 ls s3://bucket/foo', expected_rc=0)

    def test_success_rc_has_only_objects(self):
        time_utc = "2014-01-09T20:45:49.000Z"
        self.parsed_responses = [
            {
                "Contents": [
                    {
                        "Key": "foo/bar.txt",
                        "Size": 100,
                        "LastModified": time_utc,
                    }
                ]
            }
        ]
        self.run_cmd('s3 ls s3://bucket/foo', expected_rc=0)

    def test_success_rc_with_pagination(self):
        time_utc = "2014-01-09T20:45:49.000Z"
        # Pagination should not affect a successful return code of zero, even
        # if there are no results on the second page because there were
        # results in previous pages.
        self.parsed_responses = [
            {
                "CommonPrefixes": [{"Prefix": "foo/"}],
                "Contents": [
                    {
                        "Key": "foo/bar.txt",
                        "Size": 100,
                        "LastModified": time_utc,
                    }
                ],
            },
            {},
        ]
        self.run_cmd('s3 ls s3://bucket/foo', expected_rc=0)

    def test_success_rc_empty_bucket_no_key_given(self):
        # If no key has been provdided and the bucket is empty, it should
        # still return an rc of 0 since the user is not looking for an actual
        # object.
        self.parsed_responses = [{}]
        self.run_cmd('s3 ls s3://bucket', expected_rc=0)

    def test_fail_rc_no_objects_nor_prefixes(self):
        self.parsed_responses = [{}]
        self.run_cmd('s3 ls s3://bucket/foo', expected_rc=1)

    def test_human_readable_file_size(self):
        time_utc = "2014-01-09T20:45:49.000Z"
        self.parsed_responses = [
            {
                "CommonPrefixes": [],
                "Contents": [
                    {
                        "Key": "onebyte.txt",
                        "Size": 1,
                        "LastModified": time_utc,
                    },
                    {
                        "Key": "onekilobyte.txt",
                        "Size": 1024,
                        "LastModified": time_utc,
                    },
                    {
                        "Key": "onemegabyte.txt",
                        "Size": 1024**2,
                        "LastModified": time_utc,
                    },
                    {
                        "Key": "onegigabyte.txt",
                        "Size": 1024**3,
                        "LastModified": time_utc,
                    },
                    {
                        "Key": "oneterabyte.txt",
                        "Size": 1024**4,
                        "LastModified": time_utc,
                    },
                    {
                        "Key": "onepetabyte.txt",
                        "Size": 1024**5,
                        "LastModified": time_utc,
                    },
                ],
            }
        ]
        stdout, _, _ = self.run_cmd(
            's3 ls s3://bucket/ --human-readable', expected_rc=0
        )
        call_args = self.operations_called[0][1]
        # Time is stored in UTC timezone, but the actual time displayed
        # is specific to your tzinfo, so shift the timezone to your local's.
        time_local = parser.parse(time_utc).astimezone(tz.tzlocal())
        time_fmt = time_local.strftime('%Y-%m-%d %H:%M:%S')
        self.assertIn('%s     1 Byte onebyte.txt\n' % time_fmt, stdout)
        self.assertIn('%s    1.0 KiB onekilobyte.txt\n' % time_fmt, stdout)
        self.assertIn('%s    1.0 MiB onemegabyte.txt\n' % time_fmt, stdout)
        self.assertIn('%s    1.0 GiB onegigabyte.txt\n' % time_fmt, stdout)
        self.assertIn('%s    1.0 TiB oneterabyte.txt\n' % time_fmt, stdout)
        self.assertIn('%s    1.0 PiB onepetabyte.txt\n' % time_fmt, stdout)

    def test_summarize(self):
        time_utc = "2014-01-09T20:45:49.000Z"
        self.parsed_responses = [
            {
                "CommonPrefixes": [],
                "Contents": [
                    {
                        "Key": "onebyte.txt",
                        "Size": 1,
                        "LastModified": time_utc,
                    },
                    {
                        "Key": "onekilobyte.txt",
                        "Size": 1024,
                        "LastModified": time_utc,
                    },
                    {
                        "Key": "onemegabyte.txt",
                        "Size": 1024**2,
                        "LastModified": time_utc,
                    },
                    {
                        "Key": "onegigabyte.txt",
                        "Size": 1024**3,
                        "LastModified": time_utc,
                    },
                    {
                        "Key": "oneterabyte.txt",
                        "Size": 1024**4,
                        "LastModified": time_utc,
                    },
                    {
                        "Key": "onepetabyte.txt",
                        "Size": 1024**5,
                        "LastModified": time_utc,
                    },
                ],
            }
        ]
        stdout, _, _ = self.run_cmd(
            's3 ls s3://bucket/ --summarize', expected_rc=0
        )
        call_args = self.operations_called[0][1]
        # Time is stored in UTC timezone, but the actual time displayed
        # is specific to your tzinfo, so shift the timezone to your local's.
        time_local = parser.parse(time_utc).astimezone(tz.tzlocal())
        time_fmt = time_local.strftime('%Y-%m-%d %H:%M:%S')
        self.assertIn('Total Objects: 6\n', stdout)
        self.assertIn('Total Size: 1127000493261825\n', stdout)

    def test_summarize_with_human_readable(self):
        time_utc = "2014-01-09T20:45:49.000Z"
        self.parsed_responses = [
            {
                "CommonPrefixes": [],
                "Contents": [
                    {
                        "Key": "onebyte.txt",
                        "Size": 1,
                        "LastModified": time_utc,
                    },
                    {
                        "Key": "onekilobyte.txt",
                        "Size": 1024,
                        "LastModified": time_utc,
                    },
                    {
                        "Key": "onemegabyte.txt",
                        "Size": 1024**2,
                        "LastModified": time_utc,
                    },
                    {
                        "Key": "onegigabyte.txt",
                        "Size": 1024**3,
                        "LastModified": time_utc,
                    },
                    {
                        "Key": "oneterabyte.txt",
                        "Size": 1024**4,
                        "LastModified": time_utc,
                    },
                    {
                        "Key": "onepetabyte.txt",
                        "Size": 1024**5,
                        "LastModified": time_utc,
                    },
                ],
            }
        ]
        stdout, _, _ = self.run_cmd(
            's3 ls s3://bucket/ --human-readable --summarize', expected_rc=0
        )
        call_args = self.operations_called[0][1]
        # Time is stored in UTC timezone, but the actual time displayed
        # is specific to your tzinfo, so shift the timezone to your local's.
        time_local = parser.parse(time_utc).astimezone(tz.tzlocal())
        time_fmt = time_local.strftime('%Y-%m-%d %H:%M:%S')
        self.assertIn('Total Objects: 6\n', stdout)
        self.assertIn('Total Size: 1.0 PiB\n', stdout)

    def test_requester_pays(self):
        time_utc = "2014-01-09T20:45:49.000Z"
        self.parsed_responses = [
            {
                "CommonPrefixes": [],
                "Contents": [
                    {
                        "Key": "onebyte.txt",
                        "Size": 1,
                        "LastModified": time_utc,
                    },
                ],
            }
        ]
        command = 's3 ls s3://mybucket/foo/ --request-payer requester'
        self.assert_params_for_cmd(
            command,
            {
                'Bucket': 'mybucket',
                'Delimiter': '/',
                'RequestPayer': 'requester',
                'EncodingType': 'url',
                'Prefix': 'foo/',
            },
        )

    def test_requester_pays_with_no_args(self):
        time_utc = "2014-01-09T20:45:49.000Z"
        self.parsed_responses = [
            {
                "CommonPrefixes": [],
                "Contents": [
                    {
                        "Key": "onebyte.txt",
                        "Size": 1,
                        "LastModified": time_utc,
                    },
                ],
            }
        ]
        command = 's3 ls s3://mybucket/foo/ --request-payer'
        self.assert_params_for_cmd(
            command,
            {
                'Bucket': 'mybucket',
                'Delimiter': '/',
                'RequestPayer': 'requester',
                'EncodingType': 'url',
                'Prefix': 'foo/',
            },
        )

    def test_accesspoint_arn(self):
        self.parsed_responses = [self.list_objects_response(['bar.txt'])]
        arn = 'arn:aws:s3:us-west-2:123456789012:accesspoint/endpoint'
        self.run_cmd('s3 ls s3://%s' % arn, expected_rc=0)
        call_args = self.operations_called[0][1]
        self.assertEqual(call_args['Bucket'], arn)

    def test_list_buckets_uses_bucket_name_prefix(self):
        stdout, _, _ = self.run_cmd(
            's3 ls --bucket-name-prefix myprefix', expected_rc=0
        )
        call_args = self.operations_called[0][1]
        self.assertEqual(call_args['Prefix'], 'myprefix')

    def test_list_buckets_uses_bucket_region(self):
        stdout, _, _ = self.run_cmd(
            's3 ls --bucket-region us-west-1', expected_rc=0
        )
        call_args = self.operations_called[0][1]
        self.assertEqual(call_args['BucketRegion'], 'us-west-1')

    def test_list_objects_ignores_bucket_name_prefix(self):
        stdout, _, _ = self.run_cmd(
            's3 ls s3://mybucket --bucket-name-prefix myprefix', expected_rc=0
        )
        call_args = self.operations_called[0][1]
        self.assertEqual(call_args['Prefix'], '')

    def test_list_objects_ignores_bucket_region(self):
        stdout, _, _ = self.run_cmd(
            's3 ls s3://mybucket --bucket-region us-west-1', expected_rc=0
        )
        call_args = self.operations_called[0][1]
        self.assertNotIn('BucketRegion', call_args)



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
import os
import shutil
import tempfile

from awscli.testutils import (
    BaseAWSCommandParamsTest,
    BaseAWSHelpOutputTest,
    cd,
    mock,
)


class TestRunWizard(BaseAWSCommandParamsTest):
    def setUp(self):
        super(TestRunWizard, self).setUp()
        self.tempdir = tempfile.mkdtemp()
        with cd(self.tempdir):
            os.mkdir('iam')
        self.parsed_responses = [{"Roles": []}]
        self.root_dir_patch = mock.patch(
            'awscli.customizations.wizard.loader.WIZARD_SPEC_DIR', self.tempdir
        )
        self.root_dir_patch.start()

    def tearDown(self):
        super(TestRunWizard, self).tearDown()
        shutil.rmtree(self.tempdir)
        self.root_dir_patch.stop()

    def test_can_run_wizard(self):
        wizard_path = os.path.join(self.tempdir, 'iam', 'test-wizard.yml')
        with open(wizard_path, 'w') as f:
            f.write(
                'version: "0.1"\n'
                'plan:\n'
                '  start:\n'
                '    values:\n'
                '      myprefix:\n'
                '        type: static\n'
                '        value: /foo/\n'
                'execute:\n'
                '  default:\n'
                '    - type: apicall\n'
                '      operation: iam.ListRoles\n'
                '      params:\n'
                '        PathPrefix: "{myprefix}"\n'
            )
        stdout, _, _ = self.assert_params_for_cmd(
            'iam wizard test-wizard', params={'PathPrefix': '/foo/'}
        )
        self.assertEqual(self.operations_called[0][0].name, 'ListRoles')

    def test_can_run_main_wizard(self):
        wizard_path = os.path.join(self.tempdir, 'iam', '_main.yml')
        with open(wizard_path, 'w') as f:
            f.write(
                'version: "0.1"\n'
                'plan:\n'
                '  start:\n'
                '    values:\n'
                '      myprefix:\n'
                '        type: static\n'
                '        value: /foo/\n'
                'execute:\n'
                '  default:\n'
                '    - type: apicall\n'
                '      operation: iam.ListRoles\n'
                '      params: {}\n'
            )
        stdout, _, _ = self.assert_params_for_cmd(
            'iam wizard',
            params={},
        )
        self.assertEqual(self.operations_called[0][0].name, 'ListRoles')


class TestWizardHelpCommand(BaseAWSHelpOutputTest):
    def test_wizard_help_command(self):
        self.driver.main(['iam', 'wizard', 'help'])
        self.assert_contains('new-role')

    def test_wizard_subcommand_help_command(self):
        self.driver.main(['iam', 'wizard', 'new-role', 'help'])
        self.assert_contains('new-role')



# Copyright 2025 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
import pytest

from awscli.customizations.emrcontainers.base36 import Base36
from awscli.customizations.emrcontainers.constants import (
    SERVICE_ACCOUNT_NAMING,
    ServiceAccount,
)
from tests import CLIRunner, SessionStubber


@pytest.fixture
def session_stubber():
    return SessionStubber()


@pytest.fixture
def region():
    return "us-west-2"


@pytest.fixture
def cli_runner(session_stubber, region):
    cli_runner = CLIRunner(session_stubber=session_stubber)
    cli_runner.env["AWS_DEFAULT_REGION"] = region
    return cli_runner


@pytest.fixture
def account_id():
    return "123456789012"


@pytest.fixture
def cluster_name():
    return "test-cluster"


@pytest.fixture
def namespace():
    return "test"


@pytest.fixture
def role_name():
    return "myrole"


@pytest.fixture
def role_arn(account_id, role_name):
    return f"arn:aws:iam::{account_id}:role/{role_name}"


@pytest.fixture
def base36_encoded_role_name(role_name):
    return Base36().encode(role_name)


## StartJobRun expected service accounts
@pytest.fixture
def start_job_run_service_accounts(account_id, base36_encoded_role_name):
    emr_spark_components = ["client", "driver", "executor"]
    return [
        SERVICE_ACCOUNT_NAMING
        % {
            "FRAMEWORK": "spark",
            "COMPONENT": component,
            "AWS_ACCOUNT_ID": account_id,
            "BASE36_ENCODED_ROLE_NAME": base36_encoded_role_name,
        }
        for component in emr_spark_components
    ]


## InteractiveEndpoint expected service accounts
@pytest.fixture
def interactive_endpoint_service_accounts(
    account_id, base36_encoded_role_name
):
    emr_spark_components = ["jeg", "jeg-kernel", "session"]
    return [
        SERVICE_ACCOUNT_NAMING
        % {
            "FRAMEWORK": "spark",
            "COMPONENT": component,
            "AWS_ACCOUNT_ID": account_id,
            "BASE36_ENCODED_ROLE_NAME": base36_encoded_role_name,
        }
        for component in emr_spark_components
    ]


## SparkOperator expected service accounts
@pytest.fixture
def spark_operator_service_accounts(account_id, base36_encoded_role_name):
    emr_spark_operator_components = ["driver", "executor"]
    spark_operator_service_accounts = [
        ServiceAccount.SPARK_OPERATOR_SERVICE_ACCOUNT.value
    ]
    spark_operator_service_accounts.extend(
        [
            SERVICE_ACCOUNT_NAMING
            % {
                "FRAMEWORK": "spark",
                "COMPONENT": component,
                "AWS_ACCOUNT_ID": account_id,
                "BASE36_ENCODED_ROLE_NAME": base36_encoded_role_name,
            }
            for component in emr_spark_operator_components
        ]
    )
    return spark_operator_service_accounts


## FlinkOperator expected service accounts
@pytest.fixture
def flink_operator_service_accounts(account_id, base36_encoded_role_name):
    emr_flink_operator_components = ["jobmanager", "taskmanager"]
    flink_operator_service_accounts = [
        ServiceAccount.FLINK_OPERATOR_SERVICE_ACCOUNT.value
    ]
    flink_operator_service_accounts.extend(
        [
            SERVICE_ACCOUNT_NAMING
            % {
                "FRAMEWORK": "flink",
                "COMPONENT": component,
                "AWS_ACCOUNT_ID": account_id,
                "BASE36_ENCODED_ROLE_NAME": base36_encoded_role_name,
            }
            for component in emr_flink_operator_components
        ]
    )
    return flink_operator_service_accounts


## Livy expected service accounts
@pytest.fixture
def livy_service_accounts():
    return [
        ServiceAccount.LIVY_SERVICE_ACCOUNT.value,
        ServiceAccount.LIVY_SPARK_SERVICE_ACCOUNT.value,
    ]


@pytest.fixture
def create_pod_identity_association_response():
    return b'{"association": {"associationId": "a-12345678", "clusterName": "test-cluster", "namespace": "namespace", "serviceAccount":"dummy"}}'


@pytest.fixture
def describe_cluster_response():
    return b'{"cluster": {"arn": "arn:aws:eks:us-west-2:123456789012:cluster/test-cluster"}}'


@pytest.fixture
def list_pod_identity_association_response():
    return [
        b'{"associations":[{"associationId": "1", "clusterName": "test-cluster", "namespace": "namespace", "serviceAccount":"dummy1"}]}',
        b'{"associations":[{"associationId": "2", "clusterName": "test-cluster", "namespace": "namespace", "serviceAccount":"dummy2"}]}',
        b'{"associations":[{"associationId": "3", "clusterName": "test-cluster", "namespace": "namespace", "serviceAccount":"dummy3"}]}',
    ]


@pytest.fixture
def delete_pod_identity_association_response():
    return [
        b'{"association": {"association_id": "1", "clusterName": "test-cluster", "namespace": "namespace", "serviceAccount":"dummy1"}}',
        b'{"association": {"association_id": "2", "clusterName": "test-cluster", "namespace": "namespace", "serviceAccount":"dummy2"}}',
        b'{"association": {"association_id": "3", "clusterName": "test-cluster", "namespace": "namespace", "serviceAccount":"dummy3"}}',
    ]


@pytest.fixture
def create_pod_identity_association_already_exists_error_response():
    return b'{"Code": "ResourceInUseException","Message": "An error occurred (ResourceInUseException) when calling the CreatePodIdentityAssociation operation: Association already exists: a-1"}'


@pytest.fixture
def create_pod_identity_association_other_error_response():
    return b'{"Code": "InvliadRequestException","Message": "Bad Request!"}'



# Copyright 2013 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
import argparse
import datetime
import ntpath
import os
import time

import pytest
from botocore.hooks import HierarchicalEmitter
from dateutil.tz import tzlocal
from s3transfer.compat import seekable

from awscli.compat import StringIO
from awscli.customizations.exceptions import ParamValidationError
from awscli.customizations.s3.utils import (
    AppendFilter,
    BucketLister,
    NonSeekableStream,
    RequestParamsMapper,
    S3PathResolver,
    SetFileUtimeError,
    StablePriorityQueue,
    StdoutBytesWriter,
    block_unsupported_resources,
    create_warning,
    find_bucket_key,
    get_file_stat,
    guess_content_type,
    human_readable_size,
    human_readable_to_int,
    is_account_regional_namespace_bucket,
    relative_path,
    set_file_utime,
)
from awscli.testutils import mock, temporary_file, unittest


@pytest.fixture
def s3control_client():
    client = mock.MagicMock()
    client.get_access_point.return_value = {"Bucket": "mybucket"}
    client.list_multi_region_access_points.return_value = {
        "AccessPoints": [
            {"Alias": "myalias.mrap", "Regions": [{"Bucket": "mybucket"}]}
        ]
    }
    return client


@pytest.fixture
def sts_client():
    client = mock.MagicMock()
    client.get_caller_identity.return_value = {"Account": "123456789012"}
    return client


@pytest.fixture
def s3_path_resolver(s3control_client, sts_client):
    return S3PathResolver(s3control_client, sts_client)


@pytest.mark.parametrize(
    'bytes_int, expected',
    (
        (1, '1 Byte'),
        (10, '10 Bytes'),
        (1000, '1000 Bytes'),
        (1024, '1.0 KiB'),
        (1024**2, '1.0 MiB'),
        (1024**3, '1.0 GiB'),
        (1024**4, '1.0 TiB'),
        (1024**5, '1.0 PiB'),
        (1024**6, '1.0 EiB'),
        (1024**2 - 1, '1.0 MiB'),
        (1024**3 - 1, '1.0 GiB'),
    ),
)
def test_human_readable_size(bytes_int, expected):
    assert human_readable_size(bytes_int) == expected


@pytest.mark.parametrize(
    'size_str, expected',
    (
        ("1", 1),
        ("1024", 1024),
        ("1KB", 1024),
        ("1kb", 1024),
        ("1MB", 1024**2),
        ("1GB", 1024**3),
        ("1TB", 1024**4),
        ("1KiB", 1024),
        ("1kib", 1024),
        ("1MiB", 1024**2),
        ("1GiB", 1024**3),
        ("1TiB", 1024**4),
    ),
)
def test_convert_human_readable_to_int(size_str, expected):
    assert human_readable_to_int(size_str) == expected


class AppendFilterTest(unittest.TestCase):
    def test_call(self):
        parser = argparse.ArgumentParser()

        parser.add_argument(
            '--include', action=AppendFilter, nargs=1, dest='path'
        )
        parser.add_argument(
            '--exclude', action=AppendFilter, nargs=1, dest='path'
        )
        parsed_args = parser.parse_args(['--include', 'a', '--exclude', 'b'])
        self.assertEqual(
            parsed_args.path, [['--include', 'a'], ['--exclude', 'b']]
        )


class TestFindBucketKey(unittest.TestCase):
    def test_unicode(self):
        s3_path = '\u1234' + '/' + '\u5678'
        bucket, key = find_bucket_key(s3_path)
        self.assertEqual(bucket, '\u1234')
        self.assertEqual(key, '\u5678')

    def test_bucket(self):
        bucket, key = find_bucket_key('bucket')
        self.assertEqual(bucket, 'bucket')
        self.assertEqual(key, '')

    def test_bucket_with_slash(self):
        bucket, key = find_bucket_key('bucket/')
        self.assertEqual(bucket, 'bucket')
        self.assertEqual(key, '')

    def test_bucket_with_key(self):
        bucket, key = find_bucket_key('bucket/key')
        self.assertEqual(bucket, 'bucket')
        self.assertEqual(key, 'key')

    def test_bucket_with_key_and_prefix(self):
        bucket, key = find_bucket_key('bucket/prefix/key')
        self.assertEqual(bucket, 'bucket')
        self.assertEqual(key, 'prefix/key')

    def test_accesspoint_arn(self):
        bucket, key = find_bucket_key(
            'arn:aws:s3:us-west-2:123456789012:accesspoint/endpoint'
        )
        self.assertEqual(
            bucket, 'arn:aws:s3:us-west-2:123456789012:accesspoint/endpoint'
        )
        self.assertEqual(key, '')

    def test_accesspoint_arn_with_slash(self):
        bucket, key = find_bucket_key(
            'arn:aws:s3:us-west-2:123456789012:accesspoint/endpoint/'
        )
        self.assertEqual(
            bucket, 'arn:aws:s3:us-west-2:123456789012:accesspoint/endpoint'
        )
        self.assertEqual(key, '')

    def test_accesspoint_arn_with_key(self):
        bucket, key = find_bucket_key(
            'arn:aws:s3:us-west-2:123456789012:accesspoint/endpoint/key'
        )
        self.assertEqual(
            bucket, 'arn:aws:s3:us-west-2:123456789012:accesspoint/endpoint'
        )
        self.assertEqual(key, 'key')

    def test_accesspoint_arn_with_key_and_prefix(self):
        bucket, key = find_bucket_key(
            'arn:aws:s3:us-west-2:123456789012:accesspoint/endpoint/pre/key'
        )
        self.assertEqual(
            bucket, 'arn:aws:s3:us-west-2:123456789012:accesspoint/endpoint'
        )
        self.assertEqual(key, 'pre/key')

    def test_outpost_arn_with_colon(self):
        bucket, key = find_bucket_key(
            'arn:aws:s3-outposts:us-west-2:123456789012:outpost:op-12334:'
            'accesspoint:my-accesspoint'
        )
        self.assertEqual(
            bucket,
            (
                'arn:aws:s3-outposts:us-west-2:123456789012:outpost:op-12334:'
                'accesspoint:my-accesspoint'
            ),
        )
        self.assertEqual(key, '')

    def test_outpost_arn_with_colon_and_key(self):
        bucket, key = find_bucket_key(
            'arn:aws:s3-outposts:us-west-2:123456789012:outpost:op-12334:'
            'accesspoint:my-accesspoint/key'
        )
        self.assertEqual(
            bucket,
            (
                'arn:aws:s3-outposts:us-west-2:123456789012:outpost:op-12334:'
                'accesspoint:my-accesspoint'
            ),
        )
        self.assertEqual(key, 'key')

    def test_outpost_arn_with_colon_and_key_with_colon_in_name(self):
        bucket, key = find_bucket_key(
            'arn:aws:s3-outposts:us-west-2:123456789012:outpost:op-12334:'
            'accesspoint:my-accesspoint/key:name'
        )
        self.assertEqual(
            bucket,
            (
                'arn:aws:s3-outposts:us-west-2:123456789012:outpost:op-12334:'
                'accesspoint:my-accesspoint'
            ),
        )
        self.assertEqual(key, 'key:name')

    def test_outpost_arn_with_colon_and_key_with_slash_in_name(self):
        bucket, key = find_bucket_key(
            'arn:aws:s3-outposts:us-west-2:123456789012:outpost:op-12334:'
            'accesspoint:my-accesspoint/key/name'
        )
        self.assertEqual(
            bucket,
            (
                'arn:aws:s3-outposts:us-west-2:123456789012:outpost:op-12334:'
                'accesspoint:my-accesspoint'
            ),
        )
        self.assertEqual(key, 'key/name')

    def test_outpost_arn_with_colon_and_key_with_slash_and_colon_in_name(self):
        bucket, key = find_bucket_key(
            'arn:aws:s3-outposts:us-west-2:123456789012:outpost:op-12334:'
            'accesspoint:my-accesspoint/prefix/key:name'
        )
        self.assertEqual(
            bucket,
            (
                'arn:aws:s3-outposts:us-west-2:123456789012:outpost:op-12334:'
                'accesspoint:my-accesspoint'
            ),
        )
        self.assertEqual(key, 'prefix/key:name')

    def test_outpost_arn_with_slash(self):
        bucket, key = find_bucket_key(
            'arn:aws:s3-outposts:us-west-2:123456789012:outpost/op-12334/'
            'accesspoint/my-accesspoint'
        )
        self.assertEqual(
            bucket,
            (
                'arn:aws:s3-outposts:us-west-2:123456789012:outpost/op-12334/'
                'accesspoint/my-accesspoint'
            ),
        )
        self.assertEqual(key, '')

    def test_outpost_arn_with_slash_and_key(self):
        bucket, key = find_bucket_key(
            'arn:aws:s3-outposts:us-west-2:123456789012:outpost/op-12334/'
            'accesspoint/my-accesspoint/key'
        )
        self.assertEqual(
            bucket,
            (
                'arn:aws:s3-outposts:us-west-2:123456789012:outpost/op-12334/'
                'accesspoint/my-accesspoint'
            ),
        )
        self.assertEqual(key, 'key')

    def test_outpost_arn_with_slash_and_key_with_colon_in_name(self):
        bucket, key = find_bucket_key(
            'arn:aws:s3-outposts:us-west-2:123456789012:outpost/op-12334/'
            'accesspoint/my-accesspoint/key:name'
        )
        self.assertEqual(
            bucket,
            (
                'arn:aws:s3-outposts:us-west-2:123456789012:outpost/op-12334/'
                'accesspoint/my-accesspoint'
            ),
        )
        self.assertEqual(key, 'key:name')

    def test_outpost_arn_with_slash_and_key_with_slash_in_name(self):
        bucket, key = find_bucket_key(
            'arn:aws:s3-outposts:us-west-2:123456789012:outpost/op-12334/'
            'accesspoint/my-accesspoint/key/name'
        )
        self.assertEqual(
            bucket,
            (
                'arn:aws:s3-outposts:us-west-2:123456789012:outpost/op-12334/'
                'accesspoint/my-accesspoint'
            ),
        )
        self.assertEqual(key, 'key/name')

    def test_outpost_arn_with_slash_and_key_with_slash_and_colon_in_name(self):
        bucket, key = find_bucket_key(
            'arn:aws:s3-outposts:us-west-2:123456789012:outpost/op-12334/'
            'accesspoint/my-accesspoint/prefix/key:name'
        )
        self.assertEqual(
            bucket,
            (
                'arn:aws:s3-outposts:us-west-2:123456789012:outpost/op-12334/'
                'accesspoint/my-accesspoint'
            ),
        )
        self.assertEqual(key, 'prefix/key:name')


class TestBlockUnsupportedResources(unittest.TestCase):
    def test_object_lambda_arn_with_colon_raises_exception(self):
        with self.assertRaisesRegex(
            ParamValidationError, 'Use s3api commands instead'
        ):
            block_unsupported_resources(
                'arn:aws:s3-object-lambda:us-west-2:123456789012:'
                'accesspoint:my-accesspoint'
            )

    def test_object_lambda_arn_with_slash_raises_exception(self):
        with self.assertRaisesRegex(
            ParamValidationError, 'Use s3api commands instead'
        ):
            block_unsupported_resources(
                'arn:aws:s3-object-lambda:us-west-2:123456789012:'
                'accesspoint/my-accesspoint'
            )

    def test_outpost_bucket_arn_with_colon_raises_exception(self):
        with self.assertRaisesRegex(
            ParamValidationError, 'Use s3control commands instead'
        ):
            block_unsupported_resources(
                'arn:aws:s3-outposts:us-west-2:123456789012:'
                'outpost/op-0a12345678abcdefg:bucket/bucket-foo'
            )

    def test_outpost_bucket_arn_with_slash_raises_exception(self):
        with self.assertRaisesRegex(
            ParamValidationError, 'Use s3control commands instead'
        ):
            block_unsupported_resources(
                'arn:aws:s3-outposts:us-west-2:123456789012:'
                'outpost/op-0a12345678abcdefg/bucket/bucket-foo'
            )


class TestIsAccountRegionalNamespaceBucket(unittest.TestCase):
    def test_matches_standard_pattern(self):
        self.assertTrue(
            is_account_regional_namespace_bucket(
                'amzn-s3-demo-bucket-111122223333-us-west-2-an'
            )
        )

    def test_matches_different_region(self):
        self.assertTrue(
            is_account_regional_namespace_bucket(
                'my-bucket-123456789012-eu-central-1-an'
            )
        )

    def test_no_match_regular_bucket(self):
        self.assertFalse(
            is_account_regional_namespace_bucket('my-regular-bucket')
        )

    def test_no_match_missing_an_suffix(self):
        self.assertFalse(
            is_account_regional_namespace_bucket(
                'bucket-111122223333-us-west-2'
            )
        )

    def test_matches_short_bucket_name(self):
        self.assertTrue(
            is_account_regional_namespace_bucket('xyz-an')
        )

    def test_no_match_express_directory_bucket(self):
        self.assertFalse(
            is_account_regional_namespace_bucket('bucket--usw2-az1--x-s3')
        )


class TestCreateWarning(unittest.TestCase):
    def test_create_warning(self):
        path = '/foo/'
        error_message = 'There was an error'
        warning_message = create_warning(path, error_message)
        self.assertEqual(
            warning_message.message,
            'warning: Skipping file /foo/. There was an error',
        )
        self.assertFalse(warning_message.error)
        self.assertTrue(warning_message.warning)


class TestGuessContentType(unittest.TestCase):
    def test_guess_content_type(self):
        self.assertEqual(guess_content_type('foo.txt'), 'text/plain')

    def test_guess_content_type_with_no_valid_matches(self):
        self.assertEqual(guess_content_type('no-extension'), None)

    def test_guess_content_type_with_unicode_error_returns_no_match(self):
        with mock.patch('mimetypes.guess_type') as guess_type_patch:
            # This should throw a UnicodeDecodeError.
            guess_type_patch.side_effect = lambda x: b'\xe2'.decode('ascii')
            self.assertEqual(guess_content_type('foo.txt'), None)


class TestRelativePath(unittest.TestCase):
    def test_relpath_normal(self):
        self.assertEqual(
            relative_path('/tmp/foo/bar', '/tmp/foo'), '.' + os.sep + 'bar'
        )

    # We need to patch out relpath with the ntpath version so
    # we can simulate testing drives on windows.
    @mock.patch('os.path.relpath', ntpath.relpath)
    def test_relpath_with_error(self):
        # Just want to check we don't get an exception raised,
        # which is what was happening previously.
        self.assertIn(r'foo\bar', relative_path(r'c:\foo\bar'))


class TestStablePriorityQueue(unittest.TestCase):
    def test_fifo_order_of_same_priorities(self):
        a = mock.Mock()
        a.PRIORITY = 5
        b = mock.Mock()
        b.PRIORITY = 5
        c = mock.Mock()
        c.PRIORITY = 1

        q = StablePriorityQueue(maxsize=10, max_priority=20)
        q.put(a)
        q.put(b)
        q.put(c)

        # First we should get c because it's the lowest priority.
        # We're using assertIs because we want the *exact* object.
        self.assertIs(q.get(), c)
        # Then a and b are the same priority, but we should get
        # a first because it was inserted first.
        self.assertIs(q.get(), a)
        self.assertIs(q.get(), b)

    def test_queue_length(self):
        a = mock.Mock()
        a.PRIORITY = 5

        q = StablePriorityQueue(maxsize=10, max_priority=20)
        self.assertEqual(q.qsize(), 0)

        q.put(a)
        self.assertEqual(q.qsize(), 1)

        q.get()
        self.assertEqual(q.qsize(), 0)

    def test_insert_max_priority_capped(self):
        q = StablePriorityQueue(maxsize=10, max_priority=20)
        a = mock.Mock()
        a.PRIORITY = 100
        q.put(a)

        self.assertIs(q.get(), a)

    def test_priority_attr_is_missing(self):
        # If priority attr is missing, we should add it
        # to the lowest priority.
        q = StablePriorityQueue(maxsize=10, max_priority=20)
        a = object()
        b = mock.Mock()
        b.PRIORITY = 5

        q.put(a)
        q.put(b)

        self.assertIs(q.get(), b)
        self.assertIs(q.get(), a)


class TestBucketList(unittest.TestCase):
    def setUp(self):
        self.client = mock.Mock()
        self.emitter = HierarchicalEmitter()
        self.client.meta.events = self.emitter
        self.date_parser = mock.Mock()
        self.date_parser.return_value = mock.sentinel.now
        self.responses = []

    def fake_paginate(self, *args, **kwargs):
        for response in self.responses:
            self.emitter.emit('after-call.s3.ListObjectsV2', parsed=response)
        return self.responses

    def test_list_objects(self):
        now = mock.sentinel.now
        self.client.get_paginator.return_value.paginate = self.fake_paginate
        individual_response_elements = [
            {
                'LastModified': '2014-02-27T04:20:38.000Z',
                'Key': 'a',
                'Size': 1,
            },
            {
                'LastModified': '2014-02-27T04:20:38.000Z',
                'Key': 'b',
                'Size': 2,
            },
            {
                'LastModified': '2014-02-27T04:20:38.000Z',
                'Key': 'c',
                'Size': 3,
            },
        ]
        self.responses = [
            {'Contents': individual_response_elements[0:2]},
            {'Contents': [individual_response_elements[2]]},
        ]
        lister = BucketLister(self.client, self.date_parser)
        objects = list(lister.list_objects(bucket='foo'))
        self.assertEqual(
            objects,
            [
                ('foo/a', individual_response_elements[0]),
                ('foo/b', individual_response_elements[1]),
                ('foo/c', individual_response_elements[2]),
            ],
        )
        for individual_response in individual_response_elements:
            self.assertEqual(individual_response['LastModified'], now)

    def test_list_objects_passes_in_extra_args(self):
        self.client.get_paginator.return_value.paginate.return_value = [
            {
                'Contents': [
                    {
                        'LastModified': '2014-02-27T04:20:38.000Z',
                        'Key': 'mykey',
                        'Size': 3,
                    }
                ]
            }
        ]
        lister = BucketLister(self.client, self.date_parser)
        list(
            lister.list_objects(
                bucket='mybucket', extra_args={'RequestPayer': 'requester'}
            )
        )
        self.client.get_paginator.return_value.paginate.assert_called_with(
            Bucket='mybucket',
            PaginationConfig={'PageSize': None},
            RequestPayer='requester',
        )


class TestGetFileStat(unittest.TestCase):
    def test_get_file_stat(self):
        now = datetime.datetime.now(tzlocal())
        epoch_now = time.mktime(now.timetuple())
        with temporary_file('w') as f:
            f.write('foo')
            f.flush()
            os.utime(f.name, (epoch_now, epoch_now))
            size, update_time = get_file_stat(f.name)
            self.assertEqual(size, 3)
            self.assertEqual(time.mktime(update_time.timetuple()), epoch_now)

    def test_error_message(self):
        with mock.patch('os.stat', mock.Mock(side_effect=OSError('msg'))):
            with self.assertRaisesRegex(ValueError, r'myfilename\.txt'):
                get_file_stat('myfilename.txt')

    def assert_handles_fromtimestamp_error(self, error):
        patch_attribute = 'awscli.customizations.s3.utils.datetime'
        with mock.patch(patch_attribute) as datetime_mock:
            with temporary_file('w') as temp_file:
                temp_file.write('foo')
                temp_file.flush()
                datetime_mock.fromtimestamp.side_effect = error
                size, update_time = get_file_stat(temp_file.name)
                self.assertIsNone(update_time)

    def test_returns_epoch_on_invalid_timestamp(self):
        self.assert_handles_fromtimestamp_error(ValueError())

    def test_returns_epoch_on_invalid_timestamp_os_error(self):
        self.assert_handles_fromtimestamp_error(OSError())

    def test_returns_epoch_on_invalid_timestamp_overflow_error(self):
        self.assert_handles_fromtimestamp_error(OverflowError())


class TestSetsFileUtime(unittest.TestCase):
    def test_successfully_sets_utime(self):
        now = datetime.datetime.now(tzlocal())
        epoch_now = time.mktime(now.timetuple())
        with temporary_file('w') as f:
            set_file_utime(f.name, epoch_now)
            _, update_time = get_file_stat(f.name)
            self.assertEqual(time.mktime(update_time.timetuple()), epoch_now)

    def test_throws_more_relevant_error_when_errno_1(self):
        now = datetime.datetime.now(tzlocal())
        epoch_now = time.mktime(now.timetuple())
        with mock.patch('os.utime') as utime_mock:
            utime_mock.side_effect = OSError(1, '')
            with self.assertRaises(SetFileUtimeError):
                set_file_utime('not_real_file', epoch_now)

    def test_passes_through_other_os_errors(self):
        now = datetime.datetime.now(tzlocal())
        epoch_now = time.mktime(now.timetuple())
        with mock.patch('os.utime') as utime_mock:
            utime_mock.side_effect = OSError(2, '')
            with self.assertRaises(OSError):
                set_file_utime('not_real_file', epoch_now)


class TestRequestParamsMapperSSE(unittest.TestCase):
    def setUp(self):
        self.cli_params = {
            'sse': 'AES256',
            'sse_kms_key_id': 'my-kms-key',
            'sse_c': 'AES256',
            'sse_c_key': 'my-sse-c-key',
            'sse_c_copy_source': 'AES256',
            'sse_c_copy_source_key': 'my-sse-c-copy-source-key',
        }

    def test_head_object(self):
        params = {}
        RequestParamsMapper.map_head_object_params(params, self.cli_params)
        self.assertEqual(
            params,
            {
                'SSECustomerAlgorithm': 'AES256',
                'SSECustomerKey': 'my-sse-c-key',
            },
        )

    def test_put_object(self):
        params = {}
        RequestParamsMapper.map_put_object_params(params, self.cli_params)
        self.assertEqual(
            params,
            {
                'SSECustomerAlgorithm': 'AES256',
                'SSECustomerKey': 'my-sse-c-key',
                'SSEKMSKeyId': 'my-kms-key',
                'ServerSideEncryption': 'AES256',
            },
        )

    def test_get_object(self):
        params = {}
        RequestParamsMapper.map_get_object_params(params, self.cli_params)
        self.assertEqual(
            params,
            {
                'SSECustomerAlgorithm': 'AES256',
                'SSECustomerKey': 'my-sse-c-key',
            },
        )

    def test_copy_object(self):
        params = {}
        RequestParamsMapper.map_copy_object_params(params, self.cli_params)
        self.assertEqual(
            params,
            {
                'CopySourceSSECustomerAlgorithm': 'AES256',
                'CopySourceSSECustomerKey': 'my-sse-c-copy-source-key',
                'SSECustomerAlgorithm': 'AES256',
                'SSECustomerKey': 'my-sse-c-key',
                'SSEKMSKeyId': 'my-kms-key',
                'ServerSideEncryption': 'AES256',
            },
        )

    def test_create_multipart_upload(self):
        params = {}
        RequestParamsMapper.map_create_multipart_upload_params(
            params, self.cli_params
        )
        self.assertEqual(
            params,
            {
                'SSECustomerAlgorithm': 'AES256',
                'SSECustomerKey': 'my-sse-c-key',
                'SSEKMSKeyId': 'my-kms-key',
                'ServerSideEncryption': 'AES256',
            },
        )

    def test_upload_part(self):
        params = {}
        RequestParamsMapper.map_upload_part_params(params, self.cli_params)
        self.assertEqual(
            params,
            {
                'SSECustomerAlgorithm': 'AES256',
                'SSECustomerKey': 'my-sse-c-key',
            },
        )

    def test_upload_part_copy(self):
        params = {}
        RequestParamsMapper.map_upload_part_copy_params(
            params, self.cli_params
        )
        self.assertEqual(
            params,
            {
                'CopySourceSSECustomerAlgorithm': 'AES256',
                'CopySourceSSECustomerKey': 'my-sse-c-copy-source-key',
                'SSECustomerAlgorithm': 'AES256',
                'SSECustomerKey': 'my-sse-c-key',
            },
        )


class TestRequestParamsMapperChecksumAlgorithm:
    @pytest.fixture
    def cli_params(self):
        return {'checksum_algorithm': 'CRC64NVME'}

    @pytest.fixture
    def cli_params_no_algorithm(self):
        return {}

    def test_put_object(self, cli_params):
        request_params = {}
        RequestParamsMapper.map_put_object_params(request_params, cli_params)
        assert request_params == {'ChecksumAlgorithm': 'CRC64NVME'}

    def test_put_object_no_checksum(self, cli_params_no_algorithm):
        request_params = {}
        RequestParamsMapper.map_put_object_params(
            request_params, cli_params_no_algorithm
        )
        assert 'ChecksumAlgorithm' not in request_params

    def test_copy_object(self, cli_params):
        request_params = {}
        RequestParamsMapper.map_copy_object_params(request_params, cli_params)
        assert request_params == {'ChecksumAlgorithm': 'CRC64NVME'}

    def test_copy_object_no_checksum(self, cli_params_no_algorithm):
        request_params = {}
        RequestParamsMapper.map_put_object_params(
            request_params, cli_params_no_algorithm
        )
        assert 'ChecksumAlgorithm' not in request_params


class TestRequestParamsMapperChecksumMode:
    @pytest.fixture
    def cli_params(self):
        return {'checksum_mode': 'ENABLED'}

    @pytest.fixture
    def cli_params_no_checksum(self):
        return {}

    def test_get_object(self, cli_params):
        request_params = {}
        RequestParamsMapper.map_get_object_params(request_params, cli_params)
        assert request_params == {'ChecksumMode': 'ENABLED'}

    def test_get_object_no_checksums(self, cli_params_no_checksum):
        request_params = {}
        RequestParamsMapper.map_get_object_params(
            request_params, cli_params_no_checksum
        )
        assert 'ChecksumMode' not in request_params

    def test_head_object(self, cli_params):
        request_params = {}
        RequestParamsMapper.map_head_object_params(request_params, cli_params)
        assert request_params == {'ChecksumMode': 'ENABLED'}

    def test_head_object_no_checksums(self, cli_params_no_checksum):
        request_params = {}
        RequestParamsMapper.map_head_object_params(
            request_params, cli_params_no_checksum
        )
        assert 'ChecksumMode' not in request_params


class TestRequestParamsMapperRequestPayer(unittest.TestCase):
    def setUp(self):
        self.cli_params = {'request_payer': 'requester'}

    def test_head_object(self):
        params = {}
        RequestParamsMapper.map_head_object_params(params, self.cli_params)
        self.assertEqual(params, {'RequestPayer': 'requester'})

    def test_put_object(self):
        params = {}
        RequestParamsMapper.map_put_object_params(params, self.cli_params)
        self.assertEqual(params, {'RequestPayer': 'requester'})

    def test_get_object(self):
        params = {}
        RequestParamsMapper.map_get_object_params(params, self.cli_params)
        self.assertEqual(params, {'RequestPayer': 'requester'})

    def test_copy_object(self):
        params = {}
        RequestParamsMapper.map_copy_object_params(params, self.cli_params)
        self.assertEqual(params, {'RequestPayer': 'requester'})

    def test_create_multipart_upload(self):
        params = {}
        RequestParamsMapper.map_create_multipart_upload_params(
            params, self.cli_params
        )
        self.assertEqual(params, {'RequestPayer': 'requester'})

    def test_upload_part(self):
        params = {}
        RequestParamsMapper.map_upload_part_params(params, self.cli_params)
        self.assertEqual(params, {'RequestPayer': 'requester'})

    def test_upload_part_copy(self):
        params = {}
        RequestParamsMapper.map_upload_part_copy_params(
            params, self.cli_params
        )
        self.assertEqual(params, {'RequestPayer': 'requester'})

    def test_delete_object(self):
        params = {}
        RequestParamsMapper.map_delete_object_params(params, self.cli_params)
        self.assertEqual(params, {'RequestPayer': 'requester'})

    def test_list_objects(self):
        params = {}
        RequestParamsMapper.map_list_objects_v2_params(params, self.cli_params)
        self.assertEqual(params, {'RequestPayer': 'requester'})

    def test_map_get_object_tagging_params(self):
        params = {}
        RequestParamsMapper.map_get_object_tagging_params(
            params, self.cli_params
        )
        self.assertEqual(params, {'RequestPayer': 'requester'})

    def test_map_put_object_tagging_params(self):
        params = {}
        RequestParamsMapper.map_put_object_tagging_params(
            params, self.cli_params
        )
        self.assertEqual(params, {'RequestPayer': 'requester'})


class TestBytesPrint(unittest.TestCase):
    def setUp(self):
        self.stdout = mock.Mock()
        self.stdout.buffer = self.stdout

    def test_stdout_wrapper(self):
        wrapper = StdoutBytesWriter(self.stdout)
        wrapper.write(b'foo')
        self.assertTrue(self.stdout.write.called)
        self.assertEqual(self.stdout.write.call_args[0][0], b'foo')


class TestNonSeekableStream(unittest.TestCase):
    def test_can_make_stream_unseekable(self):
        fileobj = StringIO('foobar')
        self.assertTrue(seekable(fileobj))
        nonseekable_fileobj = NonSeekableStream(fileobj)
        self.assertFalse(seekable(nonseekable_fileobj))
        self.assertEqual(nonseekable_fileobj.read(), 'foobar')

    def test_can_specify_amount_for_nonseekable_stream(self):
        nonseekable_fileobj = NonSeekableStream(StringIO('foobar'))
        self.assertEqual(nonseekable_fileobj.read(3), 'foo')


class TestRequestParamsMapperNoOverwrite(unittest.TestCase):
    def test_map_put_object_params_with_no_overwrite(self):
        request_params = {}
        cli_params = {'no_overwrite': True}
        RequestParamsMapper.map_put_object_params(request_params, cli_params)
        self.assertEqual(request_params['IfNoneMatch'], '*')

    def test_map_put_object_params_without_no_overwrite(self):
        request_params = {}
        cli_params = {}
        RequestParamsMapper.map_put_object_params(request_params, cli_params)
        self.assertNotIn('IfNoneMatch', request_params)

    def test_map_copy_object_params_with_no_overwrite(self):
        request_params = {}
        cli_params = {'no_overwrite': True}
        RequestParamsMapper.map_copy_object_params(request_params, cli_params)
        self.assertEqual(request_params['IfNoneMatch'], '*')

    def test_map_copy_object_params_without_no_overwrite(self):
        request_params = {}
        cli_params = {}
        RequestParamsMapper.map_copy_object_params(request_params, cli_params)
        self.assertNotIn('IfNoneMatch', request_params)


class TestS3PathResolver:
    _BASE_ACCESSPOINT_ARN = (
        "s3://arn:aws:s3:us-west-2:123456789012:accesspoint/myaccesspoint"
    )
    _BASE_OUTPOST_ACCESSPOINT_ARN = (
        "s3://arn:aws:s3-outposts:us-east-1:123456789012:outpost"
        "/op-foo/accesspoint/myaccesspoint"
    )
    _BASE_ACCESSPOINT_ALIAS = "s3://myaccesspoint-foobar12345-s3alias"
    _BASE_OUTPOST_ACCESSPOINT_ALIAS = "s3://myaccesspoint-foobar12345--op-s3"
    _BASE_MRAP_ARN = "s3://arn:aws:s3::123456789012:accesspoint/myalias.mrap"

    @pytest.mark.parametrize(
        "path,resolved",
        [
            (_BASE_ACCESSPOINT_ARN, "s3://mybucket/"),
            (f"{_BASE_ACCESSPOINT_ARN}/", "s3://mybucket/"),
            (f"{_BASE_ACCESSPOINT_ARN}/mykey", "s3://mybucket/mykey"),
            (f"{_BASE_ACCESSPOINT_ARN}/myprefix/", "s3://mybucket/myprefix/"),
            (
                f"{_BASE_ACCESSPOINT_ARN}/myprefix/mykey",
                "s3://mybucket/myprefix/mykey",
            ),
        ],
    )
    def test_resolves_accesspoint_arn(
        self, path, resolved, s3_path_resolver, s3control_client
    ):
        resolved_paths = s3_path_resolver.resolve_underlying_s3_paths(path)
        assert resolved_paths == [resolved]
        s3control_client.get_access_point.assert_called_with(
            AccountId="123456789012", Name="myaccesspoint"
        )

    @pytest.mark.parametrize(
        "path,resolved",
        [
            (_BASE_OUTPOST_ACCESSPOINT_ARN, "s3://mybucket/"),
            (f"{_BASE_OUTPOST_ACCESSPOINT_ARN}/", "s3://mybucket/"),
            (f"{_BASE_OUTPOST_ACCESSPOINT_ARN}/mykey", "s3://mybucket/mykey"),
            (
                f"{_BASE_OUTPOST_ACCESSPOINT_ARN}/myprefix/",
                "s3://mybucket/myprefix/",
            ),
            (
                f"{_BASE_OUTPOST_ACCESSPOINT_ARN}/myprefix/mykey",
                "s3://mybucket/myprefix/mykey",
            ),
        ],
    )
    def test_resolves_outpost_accesspoint_arn(
        self, path, resolved, s3_path_resolver, s3control_client
    ):
        resolved_paths = s3_path_resolver.resolve_underlying_s3_paths(path)
        assert resolved_paths == [resolved]
        s3control_client.get_access_point.assert_called_with(
            AccountId="123456789012",
            Name=(
                "arn:aws:s3-outposts:us-east-1:123456789012:outpost"
                "/op-foo/accesspoint/myaccesspoint"
            ),
        )

    @pytest.mark.parametrize(
        "path,resolved",
        [
            (_BASE_ACCESSPOINT_ALIAS, "s3://mybucket/"),
            (f"{_BASE_ACCESSPOINT_ALIAS}/", "s3://mybucket/"),
            (f"{_BASE_ACCESSPOINT_ALIAS}/mykey", "s3://mybucket/mykey"),
            (
                f"{_BASE_ACCESSPOINT_ALIAS}/myprefix/",
                "s3://mybucket/myprefix/",
            ),
            (
                f"{_BASE_ACCESSPOINT_ALIAS}/myprefix/mykey",
                "s3://mybucket/myprefix/mykey",
            ),
        ],
    )
    def test_resolves_accesspoint_alias(
        self, path, resolved, s3_path_resolver, s3control_client, sts_client
    ):
        resolved_paths = s3_path_resolver.resolve_underlying_s3_paths(path)
        assert resolved_paths == [resolved]
        sts_client.get_caller_identity.assert_called_once_with()
        s3control_client.get_access_point.assert_called_with(
            AccountId="123456789012", Name="myaccesspoint-foobar12345-s3alias"
        )

    @pytest.mark.parametrize(
        "path",
        [
            (_BASE_OUTPOST_ACCESSPOINT_ALIAS),
            (f"{_BASE_OUTPOST_ACCESSPOINT_ALIAS}/"),
            (f"{_BASE_OUTPOST_ACCESSPOINT_ALIAS}/mykey"),
            (f"{_BASE_OUTPOST_ACCESSPOINT_ALIAS}/myprefix/"),
            (f"{_BASE_OUTPOST_ACCESSPOINT_ALIAS}/myprefix/mykey"),
        ],
    )
    def test_outpost_accesspoint_alias_raises_exception(
        self, path, s3_path_resolver
    ):
        with pytest.raises(ParamValidationError) as e:
            s3_path_resolver.resolve_underlying_s3_paths(path)
        assert "Can't resolve underlying bucket name" in str(e.value)

    @pytest.mark.parametrize(
        "path,resolved",
        [
            (_BASE_MRAP_ARN, "s3://mybucket/"),
            (f"{_BASE_MRAP_ARN}/", "s3://mybucket/"),
            (f"{_BASE_MRAP_ARN}/mykey", "s3://mybucket/mykey"),
            (f"{_BASE_MRAP_ARN}/myprefix/", "s3://mybucket/myprefix/"),
            (
                f"{_BASE_MRAP_ARN}/myprefix/mykey",
                "s3://mybucket/myprefix/mykey",
            ),
        ],
    )
    def test_resolves_mrap_arn(
        self, path, resolved, s3_path_resolver, s3control_client
    ):
        resolved_paths = s3_path_resolver.resolve_underlying_s3_paths(path)
        assert resolved_paths == [resolved]
        s3control_client.list_multi_region_access_points.assert_called_with(
            AccountId="123456789012"
        )

    @pytest.mark.parametrize(
        "path,resolved,name",
        [
            (
                f"{_BASE_ACCESSPOINT_ARN}-s3alias/mykey",
                "s3://mybucket/mykey",
                "myaccesspoint-s3alias",
            ),
            (
                f"{_BASE_OUTPOST_ACCESSPOINT_ARN}--op-s3/mykey",
                "s3://mybucket/mykey",
                f"{_BASE_OUTPOST_ACCESSPOINT_ARN[5:]}--op-s3",
            ),
        ],
    )
    def test_alias_suffixes_dont_match_accesspoint_arns(
        self, path, resolved, name, s3_path_resolver, s3control_client
    ):
        resolved_paths = s3_path_resolver.resolve_underlying_s3_paths(path)
        assert resolved_paths == [resolved]
        s3control_client.get_access_point.assert_called_with(
            AccountId="123456789012", Name=name
        )

    @pytest.mark.parametrize(
        "path,expected_has_underlying_s3_path",
        [
            (_BASE_ACCESSPOINT_ARN, True),
            (f"{_BASE_ACCESSPOINT_ARN}/mykey", True),
            (f"{_BASE_ACCESSPOINT_ARN}/myprefix/mykey", True),
            (_BASE_ACCESSPOINT_ALIAS, True),
            (_BASE_OUTPOST_ACCESSPOINT_ARN, True),
            (_BASE_OUTPOST_ACCESSPOINT_ALIAS, True),
            (_BASE_MRAP_ARN, True),
            ("s3://mybucket/", False),
            ("s3://mybucket/mykey", False),
            ("s3://mybucket/myprefix/mykey", False),
        ],
    )
    def test_has_underlying_s3_path(
        self, path, expected_has_underlying_s3_path
    ):
        has_underlying_s3_path = S3PathResolver.has_underlying_s3_path(path)
        assert has_underlying_s3_path == expected_has_underlying_s3_path



#!/usr/bin/env python
# Copyright 2013 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

import os
import shutil
import tempfile
import zipfile
from unittest import TestCase

import pytest

from awscli.customizations.cloudformation.artifact_exporter import (
    Template,
    make_zip,
)
from awscli.customizations.cloudformation.yamlhelper import yaml_dump
from awscli.testutils import skip_if_windows


class TestPackageZipFiles(TestCase):
    def setUp(self):
        self.rootdir = tempfile.mkdtemp()
        self.ziproot = os.path.join(self.rootdir, "zipcontents")

        os.mkdir(self.ziproot)

    def tearDown(self):
        shutil.rmtree(self.rootdir)

    @skip_if_windows(
        "Symlinks are not supported on Python 2.x + Windows, and require "
        "administrator privileges on Python 3.x + Windows."
    )
    def test_must_follow_symlinks(self):
        data = "hello world"
        data_file = os.path.join(self.rootdir, "data.txt")

        with open(data_file, "w") as fp:
            fp.write(data)

        # Create symlink within the zip root
        link_name = os.path.join(self.ziproot, "data-link.txt")
        os.symlink(data_file, link_name)

        # Zip up the contents of folder `ziproot` which contains the symlink
        zipfile_path = make_zip(
            os.path.join(self.rootdir, "archive"), self.ziproot
        )

        # Now verify that the zipfile includes contents of the data file we created
        myzip = zipfile.ZipFile(zipfile_path)
        # Data file should be the only file within the zip
        self.assertEqual(["data-link.txt"], myzip.namelist())
        myfile = myzip.open("data-link.txt", "r")

        # Its content should be equal the value we wrote.
        self.assertEqual(data.encode("utf-8"), myfile.read())


def _generate_template_cases():
    test_case_path = os.path.join(
        os.path.dirname(os.path.abspath(__file__)), 'deploy_templates'
    )
    cases = []
    for case in os.listdir(test_case_path):
        case_path = os.path.join(test_case_path, case)
        cases.append(
            (
                os.path.join(case_path, 'input.yml'),
                os.path.join(case_path, 'output.yml'),
            )
        )
    return cases


@pytest.mark.parametrize(
    'input_template,output_template', _generate_template_cases()
)
def test_known_templates(input_template, output_template):
    template = Template(input_template, os.getcwd(), None)
    exported = template.export()
    result = yaml_dump(exported)
    expected = open(output_template).read()

    assert result == expected, (
        '\nAcutal template:\n'
        '%s'
        '\nDiffers from expected template:\n'
        '%s' % (result, expected)
    )



# Copyright 2021 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
from botocore import model

from awscli.customizations import argrename, arguments
from awscli.testutils import unittest


class TestArgumentManipulations(unittest.TestCase):
    def setUp(self):
        self.argument_table = {}

    def test_can_rename_argument(self):
        arg = arguments.CustomArgument('foo')
        self.argument_table['foo'] = arg
        handler = argrename.rename_arg('foo', 'bar')
        handler(self.argument_table)

        self.assertIn('bar', self.argument_table)
        self.assertNotIn('foo', self.argument_table)
        self.assertEqual(arg.name, 'bar')

    def test_can_alias_an_argument(self):
        arg = arguments.CustomArgument(
            'foo',
            dest='foo',
            argument_model=model.Shape('FooArg', {'type': 'string'}),
        )
        self.argument_table['foo'] = arg
        handler = argrename.hidden_alias('foo', 'alias-name')

        handler(self.argument_table)

        self.assertIn('alias-name', self.argument_table)
        self.assertIn('foo', self.argument_table)
        self.assertEqual(self.argument_table['alias-name'].name, 'alias-name')



# Copyright 2016 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
import os
import shutil
import tempfile

from awscli.customizations.configure.writer import ConfigFileWriter
from awscli.testutils import mock, skip_if_windows, unittest


class TestConfigFileWriter(unittest.TestCase):
    def setUp(self):
        self.dirname = tempfile.mkdtemp()
        self.config_filename = os.path.join(self.dirname, 'config')
        self.writer = ConfigFileWriter()

    def tearDown(self):
        shutil.rmtree(self.dirname)

    def assert_update_config(
        self, original_config_contents, updated_data, updated_config_contents
    ):
        # Given the original_config, when it's updated with update_data,
        # it should produce updated_config_contents.
        with open(self.config_filename, 'w') as f:
            f.write(original_config_contents)
        self.writer.update_config(updated_data, self.config_filename)
        with open(self.config_filename) as f:
            new_contents = f.read()
        if new_contents != updated_config_contents:
            self.fail(
                "Config file contents do not match.\n"
                "Expected contents:\n"
                "%s\n\n"
                "Actual Contents:\n"
                "%s\n" % (updated_config_contents, new_contents)
            )

    def test_update_single_existing_value(self):
        original = '[default]\nfoo = 1\nbar = 1'
        updated = '[default]\nfoo = newvalue\nbar = 1'
        self.assert_update_config(original, {'foo': 'newvalue'}, updated)

    def test_update_value_with_square_brackets(self):
        original = '[default]\nfoo = old[value]\nbar = 1'
        updated = '[default]\nfoo = new[value]\nbar = 1'
        self.assert_update_config(original, {'foo': 'new[value]'}, updated)

    def test_update_single_existing_value_no_spaces(self):
        original = '[default]\nfoo=1\nbar=1'
        updated = '[default]\nfoo = newvalue\nbar=1'
        self.assert_update_config(original, {'foo': 'newvalue'}, updated)

    def test_update_single_new_values(self):
        expected = '[default]\nfoo = 1\nbar = 2\nbaz = newvalue\n'
        self.assert_update_config(
            '[default]\nfoo = 1\nbar = 2', {'baz': 'newvalue'}, expected
        )

    def test_handles_no_spaces(self):
        expected = '[default]\nfoo=1\nbar=2\nbaz = newvalue\n'
        self.assert_update_config(
            '[default]\nfoo=1\nbar=2', {'baz': 'newvalue'}, expected
        )

    def test_insert_values_in_middle_section(self):
        original_contents = (
            '[a]\n'
            'foo = bar\n'
            'baz = bar\n'
            '\n'
            '[b]\n'
            '\n'
            'foo = bar\n'
            '[c]\n'
            'foo = bar\n'
            'baz = bar\n'
        )
        expected_contents = (
            '[a]\n'
            'foo = bar\n'
            'baz = bar\n'
            '\n'
            '[b]\n'
            '\n'
            'foo = newvalue\n'
            '[c]\n'
            'foo = bar\n'
            'baz = bar\n'
        )
        self.assert_update_config(
            original_contents,
            {'foo': 'newvalue', '__section__': 'b'},
            expected_contents,
        )

    def test_insert_new_value_in_middle_section(self):
        original_contents = (
            '[a]\n'
            'foo = bar\n'
            '\n'
            '[b]\n'
            '\n'
            'foo = bar\n'
            '\n'
            '[c]\n'
            'foo = bar\n'
        )
        expected_contents = (
            '[a]\n'
            'foo = bar\n'
            '\n'
            '[b]\n'
            '\n'
            'foo = bar\n'
            'newvalue = newvalue\n'
            '\n'
            '[c]\n'
            'foo = bar\n'
        )
        self.assert_update_config(
            original_contents,
            {'newvalue': 'newvalue', '__section__': 'b'},
            expected_contents,
        )

    def test_new_config_file(self):
        self.assert_update_config(
            '\n', {'foo': 'value'}, '\n[default]\nfoo = value\n'
        )

    def test_section_does_not_exist(self):
        original_contents = (
            '[notdefault]\n'
            'foo = bar\n'
            'baz = bar\n'
            '\n'
            '\n'
            '\n'
            '[other "section"]\n'
            '\n'
            'foo = bar\n'
        )
        appended_contents = '[default]\n' 'foo = value\n'
        self.assert_update_config(
            original_contents,
            {'foo': 'value'},
            original_contents + appended_contents,
        )

    def test_config_file_does_not_exist(self):
        self.writer.update_config({'foo': 'value'}, self.config_filename)
        with open(self.config_filename) as f:
            new_contents = f.read()
        self.assertEqual(new_contents, '[default]\nfoo = value\n')

    @skip_if_windows("Test not valid on windows.")
    def test_permissions_on_new_file(self):
        self.writer.update_config({'foo': 'value'}, self.config_filename)
        with open(self.config_filename) as f:
            f.read()
        self.assertEqual(os.stat(self.config_filename).st_mode & 0xFFF, 0o600)

    def test_update_config_with_comments(self):
        original = '[default]\n' '#foo = 1\n' 'bar = 1\n'
        self.assert_update_config(
            original,
            {'foo': 'newvalue'},
            '[default]\n' '#foo = 1\n' 'bar = 1\n' 'foo = newvalue\n',
        )

    def test_update_config_with_commented_section(self):
        original = '#[default]\n' '[default]\n' '#foo = 1\n' 'bar = 1\n'
        self.assert_update_config(
            original,
            {'foo': 'newvalue'},
            '#[default]\n'
            '[default]\n'
            '#foo = 1\n'
            'bar = 1\n'
            'foo = newvalue\n',
        )

    def test_spaces_around_key_names(self):
        original = '[default]\n' 'foo = 1\n' 'bar = 1\n'
        self.assert_update_config(
            original,
            {'foo': 'newvalue'},
            '[default]\n' 'foo = newvalue\n' 'bar = 1\n',
        )

    def test_unquoted_profile_name(self):
        original = '[profile foobar]\n' 'foo = 1\n' 'bar = 1\n'
        self.assert_update_config(
            original,
            {'foo': 'newvalue', '__section__': 'profile foobar'},
            '[profile foobar]\n' 'foo = newvalue\n' 'bar = 1\n',
        )

    def test_double_quoted_profile_name(self):
        original = '[profile "foobar"]\n' 'foo = 1\n' 'bar = 1\n'
        self.assert_update_config(
            original,
            {'foo': 'newvalue', '__section__': 'profile foobar'},
            '[profile "foobar"]\n' 'foo = newvalue\n' 'bar = 1\n',
        )

    def test_profile_with_multiple_spaces(self):
        original = '[profile "two  spaces"]\n' 'foo = 1\n' 'bar = 1\n'
        self.assert_update_config(
            original,
            {'foo': 'newvalue', '__section__': 'profile two  spaces'},
            '[profile "two  spaces"]\n' 'foo = newvalue\n' 'bar = 1\n',
        )

    def test_nested_attributes_new_file(self):
        original = ''
        self.assert_update_config(
            original,
            {'__section__': 'default', 's3': {'signature_version': 's3v4'}},
            '[default]\n' 's3 =\n' '    signature_version = s3v4\n',
        )

    def test_add_to_nested_with_nested_in_the_middle(self):
        original = '[default]\n' 's3 =\n' '    other = foo\n' 'ec2 = bar\n'
        self.assert_update_config(
            original,
            {'__section__': 'default', 's3': {'signature_version': 'newval'}},
            '[default]\n'
            's3 =\n'
            '    other = foo\n'
            '    signature_version = newval\n'
            'ec2 = bar\n',
        )

    def test_add_to_nested_with_nested_in_the_end(self):
        original = '[default]\n' 's3 =\n' '    other = foo\n'
        self.assert_update_config(
            original,
            {'__section__': 'default', 's3': {'signature_version': 'newval'}},
            '[default]\n'
            's3 =\n'
            '    other = foo\n'
            '    signature_version = newval\n',
        )

    def test_update_nested_attribute(self):
        original = (
            '[default]\n' 's3 =\n' '    signature_version = originalval\n'
        )
        self.assert_update_config(
            original,
            {'__section__': 'default', 's3': {'signature_version': 'newval'}},
            '[default]\n' 's3 =\n' '    signature_version = newval\n',
        )

    def test_updated_nested_attribute_new_section(self):
        original = (
            '[default]\n'
            's3 =\n'
            '    other = foo\n'
            '[profile foo]\n'
            'foo = bar\n'
        )
        self.assert_update_config(
            original,
            {'__section__': 'default', 's3': {'signature_version': 'newval'}},
            '[default]\n'
            's3 =\n'
            '    other = foo\n'
            '    signature_version = newval\n'
            '[profile foo]\n'
            'foo = bar\n',
        )

    def test_update_nested_attr_no_prior_nesting(self):
        original = '[default]\n' 'foo = bar\n' '[profile foo]\n' 'foo = bar\n'
        self.assert_update_config(
            original,
            {'__section__': 'default', 's3': {'signature_version': 'newval'}},
            '[default]\n'
            'foo = bar\n'
            's3 =\n'
            '    signature_version = newval\n'
            '[profile foo]\n'
            'foo = bar\n',
        )

    def test_can_handle_empty_section(self):
        original = '[default]\n'
        self.assert_update_config(
            original,
            {'region': 'us-west-2', '__section__': 'default'},
            '[default]\n' 'region = us-west-2\n',
        )

    def test_appends_newline_on_new_section(self):
        original = '[preview]\n' 'cloudfront = true'
        self.assert_update_config(
            original,
            {'region': 'us-west-2', '__section__': 'new-section'},
            '[preview]\n'
            'cloudfront = true\n'
            '[new-section]\n'
            'region = us-west-2\n',
        )

    def test_newline_in_value_raises(self):
        with open(self.config_filename, 'w') as f:
            f.write('[default]\nfoo = bar\n')
        with self.assertRaises(ValueError):
            self.writer.update_config({'foo': 'bad\nvalue'}, self.config_filename)

    def test_carriage_return_in_value_raises(self):
        with open(self.config_filename, 'w') as f:
            f.write('[default]\nfoo = bar\n')
        with self.assertRaises(ValueError):
            self.writer.update_config({'foo': 'bad\rvalue'}, self.config_filename)

    def test_newline_in_key_raises(self):
        with open(self.config_filename, 'w') as f:
            f.write('[default]\nfoo = bar\n')
        with self.assertRaises(ValueError):
            self.writer.update_config({'bad\nkey': 'value'}, self.config_filename)

    def test_newline_in_section_name_raises(self):
        with open(self.config_filename, 'w') as f:
            f.write('[default]\nfoo = bar\n')
        with self.assertRaises(ValueError):
            self.writer.update_config(
                {'foo': 'value', '__section__': 'bad\nsection'},
                self.config_filename
            )

    def test_newline_in_nested_value_raises(self):
        with open(self.config_filename, 'w') as f:
            f.write('[default]\n')
        with self.assertRaises(ValueError):
            self.writer.update_config(
                {'__section__': 'default', 's3': {'key': 'bad\nvalue'}},
                self.config_filename
            )

    @mock.patch('awscli.customizations.configure.writer.warn_if_permissive')
    def test_check_permissions_calls_warn_if_permissive(self, mock_warn):
        with open(self.config_filename, 'w') as f:
            f.write('[default]\n')
        self.writer.update_config(
            {'__section__': 'default', 'key': 'value'},
            self.config_filename,
            check_permissions=True,
        )
        mock_warn.assert_called_once_with(self.config_filename)

    @mock.patch('awscli.customizations.configure.writer.warn_if_permissive')
    def test_no_check_permissions_by_default(self, mock_warn):
        with open(self.config_filename, 'w') as f:
            f.write('[default]\n')
        self.writer.update_config(
            {'__section__': 'default', 'key': 'value'},
            self.config_filename,
        )
        mock_warn.assert_not_called()



# Copyright 2015 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
import os
import zipfile
from contextlib import closing

from awscli.testutils import BaseAWSCommandParamsTest, FileCreator


class BaseLambdaTests(BaseAWSCommandParamsTest):
    def setUp(self):
        super(BaseLambdaTests, self).setUp()
        self.files = FileCreator()
        self.temp_file = self.files.create_file('foo', 'mycontents')
        self.zip_file = os.path.join(self.files.rootdir, 'foo.zip')
        with closing(zipfile.ZipFile(self.zip_file, 'w')) as f:
            f.write(self.temp_file)
        with open(self.zip_file, 'rb') as f:
            self.zip_file_contents = f.read()

    def tearDown(self):
        super(BaseLambdaTests, self).tearDown()
        self.files.remove_all()


class TestCreateFunction(BaseLambdaTests):
    prefix = 'lambda create-function'

    def test_create_function_with_file(self):
        cmdline = self.prefix
        cmdline += ' --function-name myfunction --runtime myruntime'
        cmdline += ' --role myrole --handler myhandler'
        cmdline += ' --zip-file fileb://%s' % self.zip_file
        result = {
            'FunctionName': 'myfunction',
            'Runtime': 'myruntime',
            'Role': 'myrole',
            'Handler': 'myhandler',
            'Code': {'ZipFile': self.zip_file_contents},
        }
        self.assert_params_for_cmd(cmdline, result)

    def test_create_function_with_code_argument(self):
        cmdline = self.prefix
        cmdline += ' --function-name myfunction --runtime myruntime'
        cmdline += ' --role myrole --handler myhandler'
        cmdline += ' --code S3Bucket=mybucket,S3Key=mykey,S3ObjectVersion=vs'
        result = {
            'FunctionName': 'myfunction',
            'Runtime': 'myruntime',
            'Role': 'myrole',
            'Handler': 'myhandler',
            'Code': {
                'S3Bucket': 'mybucket',
                'S3Key': 'mykey',
                'S3ObjectVersion': 'vs',
            },
        }
        self.assert_params_for_cmd(cmdline, result)

    def test_create_function_with_code_and_zipfile_argument(self):
        cmdline = self.prefix
        cmdline += ' --function-name myfunction --runtime myruntime'
        cmdline += ' --role myrole --handler myhandler'
        cmdline += ' --code S3Bucket=mybucket,S3Key=mykey,S3ObjectVersion=vs'
        cmdline += ' --zip-file fileb://%s' % self.zip_file
        result = {
            'FunctionName': 'myfunction',
            'Runtime': 'myruntime',
            'Role': 'myrole',
            'Handler': 'myhandler',
            'Code': {
                'S3Bucket': 'mybucket',
                'S3Key': 'mykey',
                'S3ObjectVersion': 'vs',
                'ZipFile': self.zip_file_contents,
            },
        }
        self.assert_params_for_cmd(cmdline, result)

    def test_create_function_with_zip_file_in_code_argument(self):
        cmdline = self.prefix
        cmdline += ' --function-name myfunction --runtime myruntime'
        cmdline += ' --role myrole --handler myhandler'
        cmdline += ' --code S3Bucket=mybucket,S3Key=mykey,S3ObjectVersion=vs,'
        cmdline += 'ZipFile=foo'
        stdout, stderr, rc = self.run_cmd(cmdline, expected_rc=252)
        self.assertIn(
            'ZipFile cannot be provided as part of the --code', stderr
        )

    def test_create_function_with_invalid_file_contents(self):
        cmdline = self.prefix
        cmdline += ' --function-name myfunction --runtime myruntime'
        cmdline += ' --role myrole --handler myhandler'
        cmdline += ' --zip-file filename_instead_of_contents.zip'
        stdout, stderr, rc = self.run_cmd(cmdline, expected_rc=252)
        self.assertIn('must be a zip file with the fileb:// prefix', stderr)
        # Should also give a pointer to fileb:// for them.
        self.assertIn('fileb://', stderr)

    def test_not_using_fileb_prefix(self):
        cmdline = self.prefix
        cmdline += ' --function-name myfunction --runtime myruntime'
        cmdline += ' --role myrole --handler myhandler'
        # Note file:// instead of fileb://
        cmdline += ' --zip-file file://%s' % self.zip_file
        stdout, stderr, rc = self.run_cmd(cmdline, expected_rc=252)
        # Ensure we mention fileb:// to give the user an idea of
        # where to go next.
        self.assertIn('fileb://', stderr)


class TestPublishLayerVersion(BaseLambdaTests):
    prefix = 'lambda publish-layer-version'

    def test_publish_layer_version_with_file(self):
        cmdline = self.prefix
        cmdline += ' --layer-name mylayer'
        cmdline += ' --zip-file fileb://%s' % self.zip_file
        result = {
            'LayerName': 'mylayer',
            'Content': {'ZipFile': self.zip_file_contents},
        }
        self.assert_params_for_cmd(cmdline, result)

    def test_publish_layer_version_with_content_argument(self):
        cmdline = self.prefix
        cmdline += ' --layer-name mylayer'
        cmdline += ' --content'
        cmdline += ' S3Bucket=mybucket,S3Key=mykey,S3ObjectVersion=vs'
        result = {
            'LayerName': 'mylayer',
            'Content': {
                'S3Bucket': 'mybucket',
                'S3Key': 'mykey',
                'S3ObjectVersion': 'vs',
            },
        }
        self.assert_params_for_cmd(cmdline, result)

    def test_publish_layer_version_with_content_and_zipfile_argument(self):
        cmdline = self.prefix
        cmdline += ' --layer-name mylayer'
        cmdline += ' --content'
        cmdline += ' S3Bucket=mybucket,S3Key=mykey,S3ObjectVersion=vs'
        cmdline += ' --zip-file fileb://%s' % self.zip_file
        result = {
            'LayerName': 'mylayer',
            'Content': {
                'S3Bucket': 'mybucket',
                'S3Key': 'mykey',
                'S3ObjectVersion': 'vs',
                'ZipFile': self.zip_file_contents,
            },
        }
        self.assert_params_for_cmd(cmdline, result)

    def test_publish_layer_version_with_zip_file_in_content_argument(self):
        cmdline = self.prefix
        cmdline += ' --layer-name mylayer'
        cmdline += ' --content'
        cmdline += ' S3Bucket=mybucket,S3Key=mykey,S3ObjectVersion=vs,'
        cmdline += 'ZipFile=foo'
        stdout, stderr, rc = self.run_cmd(cmdline, expected_rc=252)
        self.assertIn(
            'ZipFile cannot be provided as part of the --content', stderr
        )

    def test_publish_layer_version_with_invalid_file_contents(self):
        cmdline = self.prefix
        cmdline += ' --layer-name mylayer'
        cmdline += ' --zip-file filename_instead_of_contents.zip'
        stdout, stderr, rc = self.run_cmd(cmdline, expected_rc=252)
        self.assertIn('must be a zip file with the fileb:// prefix', stderr)
        # Should also give a pointer to fileb:// for them.
        self.assertIn('fileb://', stderr)

    def test_not_using_fileb_prefix(self):
        cmdline = self.prefix
        cmdline += ' --layer-name mylayer'
        # Note file:// instead of fileb://
        cmdline += ' --zip-file file://%s' % self.zip_file
        stdout, stderr, rc = self.run_cmd(cmdline, expected_rc=252)
        # Ensure we mention fileb:// to give the user an idea of
        # where to go next.
        self.assertIn('fileb://', stderr)


class TestUpdateFunctionCode(BaseLambdaTests):
    prefix = 'lambda update-function-code'

    def test_not_using_fileb_prefix(self):
        cmdline = self.prefix + ' --function-name foo'
        cmdline += ' --zip-file filename_instead_of_contents.zip'
        stdout, stderr, rc = self.run_cmd(cmdline, expected_rc=252)
        self.assertIn('must be a zip file with the fileb:// prefix', stderr)
        # Should also give a pointer to fileb:// for them.
        self.assertIn('fileb://', stderr)

    def test_using_fileb_prefix_succeeds(self):
        cmdline = self.prefix
        cmdline += ' --function-name myfunction'
        cmdline += ' --zip-file fileb://%s' % self.zip_file
        result = {
            'FunctionName': 'myfunction',
            'ZipFile': self.zip_file_contents,
        }
        self.assert_params_for_cmd(cmdline, result)



# Copyright 2014 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
import json
import os
import re
import time

import pytest

from awscli.testutils import aws

INTEG_TESTS_DIR = os.path.dirname(os.path.abspath(__file__))

with open(
    os.path.join(INTEG_TESTS_DIR, "smoke-tests.json")
) as definitions_file:
    SMOKE_TESTS = json.load(definitions_file)
COMMANDS = [" ".join(cmd) for cmd in SMOKE_TESTS["Commands"]]
ERROR_COMMANDS = [" ".join(cmd) for cmd in SMOKE_TESTS["ErrorCommands"]]
REGION_OVERRIDES = SMOKE_TESTS.get("RegionOverrides", {})


def _aws(command_string, max_attempts=1, delay=5, target_rc=0):
    service = command_string.split()[0]
    env = None
    if service in REGION_OVERRIDES:
        env = os.environ.copy()
        env['AWS_REGION'] = REGION_OVERRIDES[service]

    for _ in range(max_attempts - 1):
        result = aws(command_string, env_vars=env)
        if result.rc == target_rc:
            return result
        time.sleep(delay)
    return aws(command_string, env_vars=env)


@pytest.mark.parametrize('cmd', COMMANDS)
def test_can_make_success_request(cmd):
    result = _aws(cmd, max_attempts=5, delay=5, target_rc=0)
    assert result.rc == 0
    assert result.stderr == ''


@pytest.mark.parametrize('cmd', ERROR_COMMANDS)
def test_display_error_message(cmd):
    result = _aws(cmd, target_rc=254)
    assert result.rc == 254
    error_message = re.compile(
        r'An error occurred \(.+\) when calling the \w+ operation: \w+'
    )
    match = error_message.search(result.stderr)
    assert (
        match
    ), f'Error message was not displayed for command "{cmd}": {result.stderr}'



# Copyright 2022 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
import sys
from enum import Enum
from pathlib import Path

ROOT_DIR = Path(__file__).parents[2]
BUILD_DIR = ROOT_DIR / "build"

EXE_DIR = ROOT_DIR / "exe"
EXE_ASSETS_DIR = EXE_DIR / "assets"
PYINSTALLER_DIR = EXE_DIR / "pyinstaller"

# Platform specific values
IS_WINDOWS = sys.platform == "win32"
BIN_DIRNAME = "Scripts" if IS_WINDOWS else "bin"
PYTHON_EXE_NAME = "python.exe" if IS_WINDOWS else "python"
PYINSTALLER_EXE_NAME = "pyinstaller.exe" if IS_WINDOWS else "pyinstaller"
CLI_SCRIPTS = ["aws.cmd"] if IS_WINDOWS else ["aws", "aws_completer"]
LOCK_SUFFIX = "win-lock.txt" if IS_WINDOWS else "lock.txt"

# Requirements files
REQUIREMENTS_DIR = ROOT_DIR / "requirements"
BOOTSTRAP_REQUIREMENTS = REQUIREMENTS_DIR / "bootstrap.txt"
DOWNLOAD_DEPS_BOOTSTRAP = REQUIREMENTS_DIR / "download-deps" / "bootstrap.txt"
DOWNLOAD_DEPS_BOOTSTRAP_LOCK = (
    REQUIREMENTS_DIR / "download-deps" / f"bootstrap-{LOCK_SUFFIX}"
)
PORTABLE_EXE_REQUIREMENTS = REQUIREMENTS_DIR / "portable-exe-extras.txt"
PORTABLE_EXE_REQUIREMENTS_LOCK = (
    REQUIREMENTS_DIR / "download-deps" / f"portable-exe-{LOCK_SUFFIX}"
)
SYSTEM_SANDBOX_REQUIREMENTS_LOCK = (
    REQUIREMENTS_DIR / "download-deps" / f"system-sandbox-{LOCK_SUFFIX}"
)

# Auto-complete index
AC_INDEX = ROOT_DIR / "awscli" / "data" / "ac.index"

INSTALL_DIRNAME = "aws-cli"


DISTRIBUTION_SOURCE_EXE = "source-exe"
DISTRIBUTION_SOURCE_SANDBOX = "source-sandbox"

# List of .dist-info directories to include in the CLI distribution.
# See also hook-awscli.py in the pyinstaller directory.
#
# prompt_toolkit uses its own metadata to determine
# its version. So we need to bundle the package
# metadata to avoid runtime errors.
# https://github.com/aws/aws-cli/issues/9453
DIST_INFO_DIRECTORIES_TO_KEEP = ['prompt_toolkit']


class ArtifactType(Enum):
    PORTABLE_EXE = "portable-exe"
    SYSTEM_SANDBOX = "system-sandbox"



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0
import base64
import datetime
from unittest import mock

import pytest
from botocore.utils import LoginCredentialsLoader
from dateutil.tz import tzutc

from awscli.botocore.exceptions import (
    ClientError,
    LoginAuthorizationCodeError,
)
from awscli.customizations.login.utils import (
    CrossDeviceLoginTokenFetcher,
    SameDeviceLoginTokenFetcher,
)

ID_TOKEN = (
    'eyJhbGciOiJIUzI1NiIsInR5cCI6IkpXVCJ9.eyJzdWIiOiJhcm46YXdzOnN0czo6MDEyMzQ1'
    'Njc4OTAxOmFzc3VtZWQtcm9sZS9BZG1pbi9hZG1pbiIsImF1ZCI6ImFybjphd3M6c2lnbmluO'
    'jo6Y2xpL3NhbWUtZGV2aWNlIiwiaXNzIjoiaHR0cHM6Ly9zaWduaW4uYXdzLmFtYXpvbi5jb2'
    '0vc2lnbmluIiwic2Vzc2lvbl9hcm4iOiJhcm46YXdzOnN0czo6MDEyMzQ1Njc4OTAxOmFzc3V'
    'tZWQtcm9sZS9BZG1pbi9hZG1pbiIsImV4cCI6MTc2MTA5OTMxMCwiaWF0IjoxNzYxMDk4NDEw'
    'fQ.kiIJ6sOFm_keeapxWtB6u5oV-sCEoNsA4vortcNW6U4'
)

EXPECTED_DPOP_KEY = (
    '-----BEGIN EC PRIVATE KEY-----\n'
    'MHcCAQEEIHjt7c+VnkIkN6RW7QgZPFNLb/9AZEhqSYYMtwrlLb3WoAoGCCqGSM49\n'
    'AwEHoUQDQgAEv2FjRpMtADMZ4zoZxshV9chEkembgzZnXSUNe+DA8dKqXN/7qTcZ\n'
    'jYJHKIi+Rn88zUGqCJo3DWF/X+ufVfdU2g==\n'
    '-----END EC PRIVATE KEY-----\n'
)

TOKEN_RESPONSE = {
    'tokenOutput': {
        'accessToken': {
            'accessKeyId': 'access_key',
            'secretAccessKey': 'secret_access_key',
            'sessionToken': 'session_token',
        },
        'tokenType': 'aws_sigv4',
        'idToken': ID_TOKEN,
        'refreshToken': 'refresh_token',
        'expiresIn': 3600,
    }
}
UTC_NOW = datetime.datetime(2025, 9, 1, 17, 39, 33, tzinfo=tzutc())

EXPECTED_TOKEN = {
    'accessToken': {
        'accessKeyId': 'access_key',
        'secretAccessKey': 'secret_access_key',
        'sessionToken': 'session_token',
        'expiresAt': '2025-09-01T18:39:33Z',
        'accountId': '012345678901',
    },
    'tokenType': 'aws_sigv4',
    'idToken': ID_TOKEN,
    'clientId': 'arn:aws:signin:::devtools/same-device',
    'refreshToken': 'refresh_token',
    'dpopKey': EXPECTED_DPOP_KEY,
}


def fake_get_base_sign_in_uri(client):
    return 'https://foobar'


@pytest.fixture
def mock_client():
    mock_client = mock.Mock()
    mock_client.create_o_auth2_token.return_value = TOKEN_RESPONSE
    return mock_client


@pytest.fixture
def mock_auth_code_fetcher():
    return mock.Mock()


@pytest.fixture
def mock_on_pending_auth():
    return mock.Mock()


@pytest.fixture
def mock_same_device_fetcher(
    mock_client, mock_auth_code_fetcher, mock_on_pending_auth, private_key
):
    with mock.patch(
        'awscli.customizations.login.utils.secrets.choice', return_value='a'
    ):
        fetcher = SameDeviceLoginTokenFetcher(
            mock_client,
            mock_auth_code_fetcher,
            mock_on_pending_auth,
            private_key,
            base_uri_builder=fake_get_base_sign_in_uri,
        )

    return fetcher


@pytest.fixture
def mock_cross_device_fetcher(mock_client, mock_on_pending_auth, private_key):
    with mock.patch(
        'awscli.customizations.login.utils.secrets.choice', return_value='a'
    ):
        fetcher = CrossDeviceLoginTokenFetcher(
            mock_client,
            mock_on_pending_auth,
            private_key,
            base_uri_builder=fake_get_base_sign_in_uri,
        )

    return fetcher


def test_same_device_fetch_token_success(
    mock_auth_code_fetcher,
    mock_same_device_fetcher,
    mock_on_pending_auth,
    mock_client,
    fake_build_dpop_header,
):
    mock_auth_code_fetcher.redirect_uri_with_port.return_value = (
        'http://localhost:8080'
    )
    mock_auth_code_fetcher.get_auth_code_and_state.return_value = (
        'auth_code',
        str(mock_same_device_fetcher._expected_state),
    )

    with mock.patch(
        'awscli.botocore.utils.build_dpop_header',
        side_effect=fake_build_dpop_header,
    ):
        with mock.patch('datetime.datetime') as mock_dt:
            mock_dt.now.return_value = UTC_NOW
            token, session_id = mock_same_device_fetcher.fetch_token()

    assert token == EXPECTED_TOKEN
    assert session_id == 'arn:aws:sts::012345678901:assumed-role/Admin/admin'
    mock_on_pending_auth.assert_called_once()
    mock_client.create_o_auth2_token.assert_called_with(
        tokenInput={
            'clientId': 'arn:aws:signin:::devtools/same-device',
            'grantType': 'authorization_code',
            'code': 'auth_code',
            'codeVerifier': 'aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa',
            'redirectUri': 'http://localhost:8080',
        },
    )


def test_same_device_fetch_token_no_auth_code(
    mock_auth_code_fetcher, mock_same_device_fetcher
):
    mock_auth_code_fetcher.get_auth_code_and_state.return_value = (
        None,
        'state',
    )

    with pytest.raises(LoginAuthorizationCodeError):
        mock_same_device_fetcher.fetch_token()


def test_same_device_fetch_token_state_mismatch(
    mock_auth_code_fetcher, mock_same_device_fetcher
):
    mock_auth_code_fetcher.get_auth_code_and_state.return_value = (
        'auth_code',
        'wrong_state',
    )

    with pytest.raises(LoginAuthorizationCodeError):
        mock_same_device_fetcher.fetch_token()


@mock.patch(
    'awscli.customizations.login.utils.CrossDeviceLoginTokenFetcher._prompt'
)
def test_cross_device_fetch_token_success(
    mock_prompt,
    mock_cross_device_fetcher,
    mock_on_pending_auth,
    mock_client,
    fake_build_dpop_header,
):
    unencoded_verification_code = (
        f'state={mock_cross_device_fetcher._expected_state}&code=auth_code'
    )
    mock_prompt.return_value = base64.b64encode(
        unencoded_verification_code.encode('utf-8')
    )

    with mock.patch(
        'awscli.botocore.utils.build_dpop_header',
        side_effect=fake_build_dpop_header,
    ):
        with mock.patch('datetime.datetime') as mock_dt:
            mock_dt.now.return_value = UTC_NOW
            token, session_id = mock_cross_device_fetcher.fetch_token()

    expected_token = EXPECTED_TOKEN.copy()
    expected_token['clientId'] = 'arn:aws:signin:::devtools/cross-device'

    assert token == expected_token
    assert session_id == 'arn:aws:sts::012345678901:assumed-role/Admin/admin'
    mock_on_pending_auth.assert_called_once()
    mock_client.create_o_auth2_token.assert_called_with(
        tokenInput={
            'clientId': 'arn:aws:signin:::devtools/cross-device',
            'grantType': 'authorization_code',
            'code': 'auth_code',
            'codeVerifier': 'aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa',
            'redirectUri': 'https://foobar/v1/sessions/confirmation',
        },
    )


@mock.patch(
    'awscli.customizations.login.utils.CrossDeviceLoginTokenFetcher._prompt'
)
def test_cross_device_fetch_token_state_mismatch(
    mock_prompt,
    mock_cross_device_fetcher,
    fake_build_dpop_header,
):
    unencoded_verification_code = 'state=wrong_state&code=auth_code'
    mock_prompt.return_value = base64.b64encode(
        unencoded_verification_code.encode('utf-8')
    )

    with mock.patch(
        'awscli.botocore.utils.build_dpop_header',
        side_effect=fake_build_dpop_header,
    ):
        with mock.patch('datetime.datetime') as mock_dt:
            mock_dt.now.return_value = UTC_NOW
            with pytest.raises(LoginAuthorizationCodeError):
                mock_cross_device_fetcher.fetch_token()


class TestLoginCredentialsLoader:
    def test_save_and_load_token(self):
        loader = LoginCredentialsLoader()
        token = {'access_token': 'token_123'}
        session_name = 'login_test'

        loader.save_token(session_name, token)

        loaded = loader.load_token(session_name)
        assert loaded == token

    def test_load_token_not_exists(self):
        loader = LoginCredentialsLoader()
        assert loader.load_token('nonexistant_session') is None


@mock.patch('awscli.customizations.login.utils.register_feature_id')
def test_same_device_fetcher_registers_login_same_device(
    mock_register, mock_auth_code_fetcher, mock_same_device_fetcher
):
    mock_auth_code_fetcher.redirect_uri_with_port.return_value = (
        'http://localhost:8080'
    )
    mock_auth_code_fetcher.get_auth_code_and_state.return_value = (
        'auth_code',
        str(mock_same_device_fetcher._expected_state),
    )

    with mock.patch.object(
        mock_same_device_fetcher, '_exchange_auth_code_for_access_token'
    ):
        mock_same_device_fetcher.fetch_token()

    mock_register.assert_called_with('LOGIN_SAME_DEVICE')


@mock.patch('awscli.customizations.login.utils.register_feature_id')
@mock.patch(
    'awscli.customizations.login.utils.CrossDeviceLoginTokenFetcher._prompt'
)
def test_cross_device_fetcher_registers_login_cross_device(
    mock_prompt, mock_register, mock_cross_device_fetcher
):
    import base64

    unencoded = (
        f'state={mock_cross_device_fetcher._expected_state}&code=auth_code'
    )
    mock_prompt.return_value = base64.b64encode(unencoded.encode('utf-8'))

    with mock.patch.object(
        mock_cross_device_fetcher, '_exchange_auth_code_for_access_token'
    ):
        mock_cross_device_fetcher.fetch_token()

    mock_register.assert_called_with('LOGIN_CROSS_DEVICE')


def test_same_device_fetch_token_expired_auth_code(
    mock_auth_code_fetcher, mock_same_device_fetcher
):
    mock_auth_code_fetcher.redirect_uri_with_port.return_value = (
        'http://localhost:8080'
    )
    mock_auth_code_fetcher.get_auth_code_and_state.return_value = (
        'expired_code',
        str(mock_same_device_fetcher._expected_state),
    )

    error_response = {
        'Error': {'Code': 'AccessDeniedException'},
        'error': 'AUTHCODE_EXPIRED',
    }
    mock_same_device_fetcher._client.create_o_auth2_token.side_effect = (
        ClientError(error_response, 'CreateOAuth2Token')
    )
    mock_same_device_fetcher._client.exceptions.AccessDeniedException = (
        ClientError
    )

    with pytest.raises(
        LoginAuthorizationCodeError, match="expired authorization code"
    ):
        mock_same_device_fetcher.fetch_token()



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
from decimal import Decimal

from awscli.customizations.dynamodb.extractor import AttributeExtractor
from awscli.testutils import unittest


class FakeParser:
    def __init__(self, parsed_result):
        self.parsed_result = parsed_result

    def parse(self, expression):
        return self.parsed_result


class TestExtractor(unittest.TestCase):
    def test_extract_identifier(self):
        parsed_result = {'type': 'identifier', 'value': 'spam'}
        parser = FakeParser(parsed_result)
        extractor = AttributeExtractor(parser)
        result = extractor.extract('spam')
        expected = {
            'expression': '#n0',
            'identifiers': {'#n0': 'spam'},
            'values': {},
            'substitution_count': 1,
        }
        self.assertEqual(result, expected)

    def test_extract_string(self):
        parsed_result = {'type': 'literal', 'value': 'spam'}
        parser = FakeParser(parsed_result)
        extractor = AttributeExtractor(parser)
        result = extractor.extract('"spam"')
        expected = {
            'expression': ':n0',
            'identifiers': {},
            'values': {':n0': 'spam'},
            'substitution_count': 1,
        }
        self.assertEqual(result, expected)

    def test_extract_bytes(self):
        parsed_result = {'type': 'literal', 'value': b'spam'}
        parser = FakeParser(parsed_result)
        extractor = AttributeExtractor(parser)
        result = extractor.extract('b"spam"')
        expected = {
            'expression': ':n0',
            'identifiers': {},
            'values': {':n0': b'spam'},
            'substitution_count': 1,
        }
        self.assertEqual(result, expected)

    def test_extract_number(self):
        parsed_result = {'type': 'literal', 'value': 7}
        parser = FakeParser(parsed_result)
        extractor = AttributeExtractor(parser)
        result = extractor.extract("7")
        expected = {
            'expression': ':n0',
            'identifiers': {},
            'values': {':n0': 7},
            'substitution_count': 1,
        }
        self.assertEqual(result, expected)

    def test_set_index_offset(self):
        parsed_result = {'type': 'identifier', 'value': 'spam'}
        parser = FakeParser(parsed_result)
        extractor = AttributeExtractor(parser)
        result = extractor.extract('spam', 5)
        expected = {
            'expression': '#n5',
            'identifiers': {'#n5': 'spam'},
            'values': {},
            'substitution_count': 1,
        }
        self.assertEqual(result, expected)

    def test_represent_comparator(self):
        parsed_result = {
            'type': 'comparator',
            'value': 'eq',
            'children': [
                {'type': 'identifier', 'value': 'spam'},
                {'type': 'literal', 'value': 7},
            ],
        }
        parser = FakeParser(parsed_result)
        extractor = AttributeExtractor(parser)
        result = extractor.extract('spam = 7')
        expected = {
            'expression': '#n0 = :n1',
            'identifiers': {'#n0': 'spam'},
            'values': {':n1': 7},
            'substitution_count': 2,
        }
        self.assertEqual(result, expected)

    def test_represent_or(self):
        parsed_result = {
            'type': 'or_expression',
            'children': [
                {
                    'type': 'comparator',
                    'value': 'eq',
                    'children': [
                        {'type': 'identifier', 'value': 'spam'},
                        {'type': 'literal', 'value': 7},
                    ],
                },
                {
                    'type': 'comparator',
                    'value': 'eq',
                    'children': [
                        {'type': 'identifier', 'value': 'eggs'},
                        {'type': 'literal', 'value': 6},
                    ],
                },
            ],
        }
        parser = FakeParser(parsed_result)
        extractor = AttributeExtractor(parser)
        result = extractor.extract('spam = 7 or eggs = 6')
        expected = {
            'expression': '#n0 = :n1 OR #n2 = :n3',
            'identifiers': {'#n0': 'spam', '#n2': 'eggs'},
            'values': {':n1': 7, ':n3': 6},
            'substitution_count': 4,
        }
        self.assertEqual(result, expected)

    def test_represent_and(self):
        parsed_result = {
            'type': 'and_expression',
            'children': [
                {
                    'type': 'comparator',
                    'value': 'eq',
                    'children': [
                        {'type': 'identifier', 'value': 'spam'},
                        {'type': 'literal', 'value': 7},
                    ],
                },
                {
                    'type': 'comparator',
                    'value': 'eq',
                    'children': [
                        {'type': 'identifier', 'value': 'eggs'},
                        {'type': 'literal', 'value': 6},
                    ],
                },
            ],
        }
        parser = FakeParser(parsed_result)
        extractor = AttributeExtractor(parser)
        result = extractor.extract('spam = 7 and eggs = 6')
        expected = {
            'expression': '#n0 = :n1 AND #n2 = :n3',
            'identifiers': {'#n0': 'spam', '#n2': 'eggs'},
            'values': {':n1': 7, ':n3': 6},
            'substitution_count': 4,
        }
        self.assertEqual(result, expected)

    def test_represent_in(self):
        parsed_result = {
            'type': 'in_expression',
            'children': [
                {'type': 'identifier', 'value': 'spam'},
                {
                    'type': 'sequence',
                    'children': [
                        {'type': 'literal', 'value': 1},
                        {'type': 'literal', 'value': 2},
                    ],
                },
            ],
        }
        parser = FakeParser(parsed_result)
        extractor = AttributeExtractor(parser)
        result = extractor.extract('spam IN (2, 3)')
        expected = {
            'expression': '#n0 IN (:n1, :n2)',
            'identifiers': {'#n0': 'spam'},
            'values': {':n1': 1, ':n2': 2},
            'substitution_count': 3,
        }
        self.assertEqual(result, expected)

    def test_represent_not(self):
        parsed_result = {
            'type': 'not_expression',
            'children': [
                {
                    'type': 'comparator',
                    'value': 'ne',
                    'children': [
                        {'type': 'identifier', 'value': 'spam'},
                        {'type': 'literal', 'value': 7},
                    ],
                },
            ],
        }
        parser = FakeParser(parsed_result)
        extractor = AttributeExtractor(parser)
        result = extractor.extract('NOT spam <> 7')
        expected = {
            'expression': 'NOT #n0 <> :n1',
            'identifiers': {'#n0': 'spam'},
            'values': {':n1': 7},
            'substitution_count': 2,
        }
        self.assertEqual(result, expected)

    def test_represent_subexpression(self):
        parsed_result = {
            'type': 'subexpression',
            'children': [
                {
                    'type': 'comparator',
                    'value': 'lte',
                    'children': [
                        {'type': 'identifier', 'value': 'spam'},
                        {'type': 'literal', 'value': 7},
                    ],
                },
            ],
        }
        parser = FakeParser(parsed_result)
        extractor = AttributeExtractor(parser)
        result = extractor.extract('( spam <= 7 )')
        expected = {
            'expression': '( #n0 <= :n1 )',
            'identifiers': {'#n0': 'spam'},
            'values': {':n1': 7},
            'substitution_count': 2,
        }
        self.assertEqual(result, expected)

    def test_represent_between(self):
        parsed_result = {
            'type': 'between_expression',
            'children': [
                {'type': 'identifier', 'value': 'spam'},
                {'type': 'literal', 'value': 1},
                {'type': 'literal', 'value': 2},
            ],
        }
        parser = FakeParser(parsed_result)
        extractor = AttributeExtractor(parser)
        result = extractor.extract('spam between 1 and 2')
        expected = {
            'expression': '#n0 BETWEEN :n1 AND :n2',
            'identifiers': {'#n0': 'spam'},
            'values': {':n1': 1, ':n2': 2},
            'substitution_count': 3,
        }
        self.assertEqual(result, expected)

    def test_represent_function(self):
        parsed_result = {
            'type': 'function',
            'value': 'myfunction',
            'children': [
                {'type': 'identifier', 'value': 'spam'},
                {'type': 'literal', 'value': 1},
            ],
        }
        parser = FakeParser(parsed_result)
        extractor = AttributeExtractor(parser)
        result = extractor.extract('myfunction(1, 2)')
        expected = {
            'expression': 'myfunction(#n0, :n1)',
            'identifiers': {'#n0': 'spam'},
            'values': {':n1': 1},
            'substitution_count': 2,
        }
        self.assertEqual(result, expected)

    def test_represent_path_identifier(self):
        parsed_result = {
            'type': 'path_identifier',
            'children': [
                {'type': 'identifier', 'value': 'foo', 'children': []},
                {'type': 'identifier', 'value': 'bar', 'children': []},
            ],
        }
        parser = FakeParser(parsed_result)
        extractor = AttributeExtractor(parser)
        result = extractor.extract('foo.bar')
        expected = {
            'expression': '#n0.#n1',
            'identifiers': {'#n0': 'foo', '#n1': 'bar'},
            'values': {},
            'substitution_count': 2,
        }
        self.assertEqual(result, expected)

    def test_represent_index_identifier(self):
        parsed_result = {
            'type': 'index_identifier',
            'value': Decimal(0),
            'children': [
                {'type': 'identifier', 'value': 'foo', 'children': []},
            ],
        }
        parser = FakeParser(parsed_result)
        extractor = AttributeExtractor(parser)
        result = extractor.extract('foo[0]')
        expected = {
            'expression': '#n0[0]',
            'identifiers': {'#n0': 'foo'},
            'values': {},
            'substitution_count': 1,
        }
        self.assertEqual(result, expected)

    def test_represent_dotted_index_identifier(self):
        parsed_result = {
            'type': 'path_identifier',
            'children': [
                {
                    'type': 'index_identifier',
                    'value': Decimal(0),
                    'children': [
                        {'type': 'identifier', 'value': 'foo', 'children': []},
                    ],
                },
                {'type': 'identifier', 'value': 'bar', 'children': []},
            ],
        }
        parser = FakeParser(parsed_result)
        extractor = AttributeExtractor(parser)
        result = extractor.extract('foo[0].bar')
        expected = {
            'expression': '#n0[0].#n1',
            'identifiers': {'#n0': 'foo', '#n1': 'bar'},
            'values': {},
            'substitution_count': 2,
        }
        self.assertEqual(result, expected)
